# pylint: disable=too-many-function-args, invalid-name
""" COBRE ICA dataset loading script"""
import h5py
import numpy as np
import pandas as pd

from pathlib import Path
DATA_ROOT = Path().joinpath("../assets/data").resolve()


def load_data(
    dataset_path: str = DATA_ROOT.joinpath("cobre/COBRE_AllData.h5"),
    indices_path: str = DATA_ROOT.joinpath("ICA_correct_order.csv"),
    labels_path: str = DATA_ROOT.joinpath("cobre/labels_COBRE.csv"),
):
    """
    Return COBRE data

    Input:
    dataset_path: str = DATA_ROOT.joinpath("cobre/COBRE_AllData.h5")
    - path to the dataset
    indices_path: str = DATA_ROOT.joinpath("ICA_correct_order.csv")
    - path to correct indices/components
    labels_path: str = DATA_ROOT.joinpath("cobre/labels_COBRE.csv")
    - path to labels

    Output:
    data, labels
    """

    # get data and reshape to correct format
    hf = h5py.File(dataset_path, "r")
    data = hf.get("COBRE_dataset")
    data = np.array(data)

    num_subjects = data.shape[0]
    num_components = 100
    data = data.reshape(num_subjects, num_components, -1)

    # get correct indices/components
    indices = pd.read_csv(indices_path, header=None)
    idx = indices[0].values - 1
    data = data[:, idx, :]

    # get labels
    labels = pd.read_csv(labels_path, header=None)
    labels = labels.values.flatten().astype("int") - 1

    data = np.swapaxes(data, 1, 2)
    # data.shape = [n_samples, time_length, feature_size]
    # 157 - sessions - data.shape[0]
    # 140 - time points - data.shape[1]
    # 53 - components - data.shape[2]

    return data, labels



import torch
from importlib.resources import files, as_file

def load_packaged_weights(rel_path):
    """
    rel_path like 'weights/whole_milc.pth' inside ml4fmri.data
    Returns a loaded object/state_dict with torch.load.
    """
    res = files("ml4fmri.data").joinpath(rel_path)
    # as_file gives you a real path even when installed from a zip wheel
    with as_file(res) as p:
        return torch.load(p, map_location="cpu")


import pandas as pd
from ml4fmri.cvreport import cvbench

def test_cvbench_smoke(toy_data):
    X, y = toy_data
    # Use your discovery inside cvbench; 'lite' defaults to a small set (e.g., meanMLP)
    rep = cvbench(X, y, models="lite", n_folds=2, val_ratio=0.2, random_state=0)
    tdf = rep.get_test_dataframe()
    trn = rep.get_train_dataframe()
    assert isinstance(tdf, pd.DataFrame) and not tdf.empty
    assert isinstance(trn, pd.DataFrame) and "epoch" in trn.columns

    # Plotting shouldn't auto-display when show=False
    out = rep.plot_scores(show=False);      assert out is not None
    figs = rep.plot_training_curves(show=False);  assert isinstance(figs, tuple) and len(figs) == 2



# tests/conftest.py
import os
import sys
from pathlib import Path
import numpy as np
import pytest

# Ensure repo src/ is importable during tests (run pytest from repo root)
SRC = Path(__file__).resolve().parents[1] / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

# Make MPS behave if present; harmless elsewhere
os.environ.setdefault("PYTORCH_ENABLE_MPS_FALLBACK", "1")

@pytest.fixture(scope="session")
def toy_data():
    """
    Deterministic, balanced binary dataset.
    We *know* where each class lives:
      - y[0:Z]   -> class 0
      - y[Z:B]   -> class 1
    """
    rng = np.random.default_rng(0)
    B, T, D = 24, 30, 10
    Z = B // 2  # zeros first, ones second

    y = np.zeros(B, dtype=np.int64)
    y[Z:] = 1

    X = rng.normal(size=(B, T, D)).astype("float32")
    # Inject signal into class 1
    X[Z:, :, :5] += 0.4
    X[Z:, 10:24, :] -= 0.2

    return X, y  # numpy arrays

@pytest.fixture(scope="session")
def dims(toy_data):
    X, y = toy_data
    _, _, D = X.shape
    C = int(np.unique(y).size)
    return D, C

@pytest.fixture(scope="session")
def toy_splits(toy_data):
    """
    Provide deterministic, balanced indices for train/val/test,
    using the known label layout from toy_data().
    Layout:
      zeros = [0:Z), ones = [Z:B)
    We take:
      train: 6 zeros + 6 ones
      val:   3 zeros + 3 ones
      test:  4 zeros + 4 ones
    (total uses 12+6+8 = 26 indices, but B=24; adjust to fit cleanly)
    """
    X, y = toy_data
    B = y.shape[0]
    Z = B // 2  # split point

    zeros = np.arange(0, Z)
    ones  = np.arange(Z, B)

    # choose counts (must sum <= class counts)
    n_tr, n_va, n_te = 6, 3, 3  # per-class counts

    tr_idx = np.concatenate([zeros[:n_tr], ones[:n_tr]])
    va_idx = np.concatenate([zeros[n_tr:n_tr+n_va], ones[n_tr:n_tr+n_va]])
    te_idx = np.concatenate([zeros[n_tr+n_va:n_tr+n_va+n_te],
                              ones[n_tr+n_va:n_tr+n_va+n_te]])

    # shuffle inside each split to mix classes a bit (deterministic)
    rng = np.random.default_rng(1)
    rng.shuffle(tr_idx); rng.shuffle(va_idx); rng.shuffle(te_idx)

    splits = {
        "train": tr_idx,
        "val": va_idx,
        "test": te_idx,
    }
    return splits



# pylint: disable=invalid-name, missing-function-docstring, missing-class-docstring, unused-argument, too-few-public-methods, no-member, too-many-arguments, line-too-long, too-many-instance-attributes
"""
BNT model module

Works with FNC data.

@inproceedings{bnt,
    title={Brain Network Transformer},
    author={Xuan Kan and Wei Dai and Hejie Cui and Zilong Zhang and Ying Guo and Carl Yang},
    booktitle={Advances in Neural Information Processing Systems},
    editor={Alice H. Oh and Alekh Agarwal and Danielle Belgrave and Kyunghyun Cho},
    year={2022},
    url={https://openreview.net/forum?id=1cJ1cbA6NLN}
}
"""


import numpy as np
import math
from torch.nn.functional import softmax
from torch.nn import Parameter, TransformerEncoderLayer, functional as F
from torch import nn, optim
import torch

from types import SimpleNamespace

from .helper_functions import basic_ce_loss, basic_handle_batch, basic_dataloader, BasicTrainer

class BNT(nn.Module):
    """
    FNC MODEL
    BrainNetworkTransformer for fMRI data from https://doi.org/10.48550/arXiv.2210.06681
    Original implementation: https://github.com/Wayfear/BrainNetworkTransformer
    """
    def __init__(self,
                 input_size,
                 output_size,
        ):
        super().__init__()

        self.lr = 1e-4  # learning rate used in the paper. Defined like this in every model for reference.
        self.weight_decay = 1e-4  # weight decay used by optimizer

        model_cfg = {
            "sizes": [360, 100],  # Note: The input node size should not be included here
            "pooling": [False, True],
            "pos_encoding": "none",  # 'identity', 'none'
            "orthogonal": True,
            "freeze_center": True,
            "project_assignment": True,
            "pos_embed_dim": 360,
            # data shape
            "node_sz": input_size,
            "node_feature_sz": input_size,
            "output_size": output_size,
            # optimizer
            "optimizer": {"lr": 1e-4, "weight_decay": 1e-4},
        }

        def dict_to_namespace(d):
            """Convert nested dict to nested SimpleNamespace"""
            if isinstance(d, dict):
                return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
            return d
        
        model_cfg = dict_to_namespace(model_cfg)

        self.attention_list = nn.ModuleList()
        forward_dim = model_cfg.node_feature_sz

        self.pos_encoding = model_cfg.pos_encoding
        if self.pos_encoding == "identity":
            self.node_identity = nn.Parameter(
                torch.zeros(model_cfg.node_sz, model_cfg.pos_embed_dim),
                requires_grad=True,
            )
            forward_dim = model_cfg.node_sz + model_cfg.pos_embed_dim
            nn.init.kaiming_normal_(self.node_identity)

        sizes = model_cfg.sizes
        sizes[0] = model_cfg.node_sz
        in_sizes = [model_cfg.node_sz] + sizes[:-1]
        do_pooling = model_cfg.pooling
        self.do_pooling = do_pooling

        self.attention_list = nn.ModuleList()
        curr_feat = forward_dim  # may be 53 or 53+pos_embed_dim

        for index, size in enumerate(sizes):
            enc = TransPoolingEncoder(
                input_feature_size=curr_feat,        # previous stage's output feature width
                input_node_num=in_sizes[index],
                hidden_size=1024,
                output_node_num=size,
                pooling=do_pooling[index],
                orthogonal=model_cfg.orthogonal,
                freeze_center=model_cfg.freeze_center,
                project_assignment=model_cfg.project_assignment,
                n_head=4,                            # keep fixed heads
            )
            self.attention_list.append(enc)
            curr_feat = enc.out_feature_size        # <- adopt padded width for next stage

            self.final_feature_size = curr_feat
            self.dim_reduction = nn.Sequential(
                nn.Linear(self.final_feature_size, 8), nn.LeakyReLU(inplace=True)
            )

        self.fc = nn.Sequential(
            nn.Linear(8 * sizes[-1], 256),
            nn.LeakyReLU(),
            nn.Linear(256, 32),
            nn.LeakyReLU(),
            nn.Linear(32, model_cfg.output_size),
        )

    def get_attention_weights(self):
        return [atten.get_attention_weights() for atten in self.attention_list]

    def get_cluster_centers(self) -> torch.Tensor:
        """
        Get the cluster centers, as computed by the encoder.

        :return: [number of clusters, hidden dimension] Tensor of dtype float
        """
        return self.dec.get_cluster_centers()

    def loss(self, assignments):
        """
        Compute KL loss for the given assignments. Note that not all encoders contain a pooling layer.
        Inputs: assignments: [batch size, number of clusters]
        Output: KL loss
        """
        decs = list(filter(lambda x: x.is_pooling_enabled(), self.attention_list))
        assignments = list(filter(lambda x: x is not None, assignments))
        loss_all = None

        for index, assignment in enumerate(assignments):
            if loss_all is None:
                loss_all = decs[index].loss(assignment)
            else:
                loss_all += decs[index].loss(assignment)
        return loss_all
    
    def forward(self, node_feature: torch.tensor):
        (
            bz,
            _,
            _,
        ) = node_feature.shape

        if self.pos_encoding == "identity":
            pos_emb = self.node_identity.expand(bz, *self.node_identity.shape)
            node_feature = torch.cat([node_feature, pos_emb], dim=-1)

        assignments = []

        for _, atten in enumerate(self.attention_list):
            node_feature, assignment = atten(node_feature)
            assignments.append(assignment)

        node_feature = self.dim_reduction(node_feature)

        node_feature = node_feature.reshape((bz, -1))

        logits = self.fc(node_feature)
        return logits, {"logits": logits}

    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """

        loss, loss_log = basic_ce_loss(loss_load["logits"], targets)
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr

        optimizer = optim.Adam(
            self.parameters(),
            lr=lr,
            weight_decay=self.weight_decay,
        )
        return optimizer

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        return basic_dataloader(data,
                                labels,
                                type="FNC",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """

        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs

class TransPoolingEncoder(nn.Module):
    """
    Transformer encoder with Pooling mechanism.
    Input : (B, input_node_num, input_feature_size)
    Output: (B, output_node_num, eff_input_size)  # NOTE: eff width (padded) is kept
    """
    def __init__(self,
                 input_feature_size,
                 input_node_num,
                 hidden_size,
                 output_node_num,
                 pooling=True,
                 orthogonal=True,
                 freeze_center=False,
                 project_assignment=True,
                 n_head=4):
        super().__init__()
        self.n_head = n_head
        self.input_size = input_feature_size
        self.eff_input_size = math.ceil(self.input_size / n_head) * n_head  # padded width
        self.pad = self.eff_input_size - self.input_size
        self.out_feature_size = self.eff_input_size  # expose for downstream

        # Run transformer at the padded width
        self.transformer = InterpretableTransformerEncoder(
            d_model=self.eff_input_size,      # ← was input_feature_size
            nhead=n_head,
            dim_feedforward=hidden_size,
            batch_first=True,
        )

        self.pooling = pooling
        if pooling:
            encoder_hidden_size = 32
            # DEC works at the padded width, too
            self.encoder = nn.Sequential(
                nn.Linear(self.eff_input_size * input_node_num, encoder_hidden_size),
                nn.LeakyReLU(inplace=True),
                nn.Linear(encoder_hidden_size, encoder_hidden_size),
                nn.LeakyReLU(inplace=True),
                nn.Linear(encoder_hidden_size, self.eff_input_size * input_node_num),
            )
            self.dec = DEC(
                cluster_number=output_node_num,
                hidden_dimension=self.eff_input_size,  # ← was input_feature_size
                encoder=self.encoder,
                orthogonal=orthogonal,
                freeze_center=freeze_center,
                project_assignment=project_assignment,
            )

    def is_pooling_enabled(self):
        return self.pooling

    def forward(self, x):  # x: (B, N, input_size)
        if self.pad:
            # pad last dim: (left=0, right=self.pad)
            x = F.pad(x, (0, self.pad))                 # (B, N, eff_input_size)
        x = self.transformer(x)                         # (B, N, eff_input_size)
        if self.pooling:
            x, assignment = self.dec(x)                 # (B, out_nodes, eff_input_size)
            return x, assignment
        return x, None

    def get_attention_weights(self):
        return self.transformer.get_attention_weights()

    def loss(self, assignment):
        return self.dec.loss(assignment)

    
class DEC(nn.Module):
    def __init__(
        self,
        cluster_number: int,
        hidden_dimension: int,
        encoder: torch.nn.Module,
        alpha: float = 1.0,
        orthogonal=True,
        freeze_center=True,
        project_assignment=True,
    ):
        """
        Module which holds all the moving parts of the DEC algorithm, as described in
        Xie/Girshick/Farhadi; this includes the AutoEncoder stage and the ClusterAssignment stage.

        :param cluster_number: number of clusters
        :param hidden_dimension: hidden dimension, output of the encoder
        :param encoder: encoder to use
        :param alpha: parameter representing the degrees of freedom in the t-distribution, default 1.0
        """
        super(DEC, self).__init__()
        self.encoder = encoder
        self.hidden_dimension = hidden_dimension
        self.cluster_number = cluster_number
        self.alpha = alpha
        self.assignment = ClusterAssignment(
            cluster_number,
            self.hidden_dimension,
            alpha,
            orthogonal=orthogonal,
            freeze_center=freeze_center,
            project_assignment=project_assignment,
        )

        self.loss_fn = nn.KLDivLoss(reduction='batchmean')

    def forward(self, batch: torch.Tensor):
        """
        Compute the cluster assignment using the ClusterAssignment after running the batch
        through the encoder part of the associated AutoEncoder module.

        :param batch: [batch size, embedding dimension] FloatTensor
        :return: [batch size, number of clusters] FloatTensor
        """
        node_num = batch.size(1)
        batch_size = batch.size(0)

        # [batch size, embedding dimension]
        flattened_batch = batch.view(batch_size, -1)
        encoded = self.encoder(flattened_batch)
        # [batch size * node_num, hidden dimension]
        encoded = encoded.view(batch_size * node_num, -1)
        # [batch size * node_num, cluster_number]
        assignment = self.assignment(encoded)
        # [batch size, node_num, cluster_number]
        assignment = assignment.view(batch_size, node_num, -1)
        # [batch size, node_num, hidden dimension]
        encoded = encoded.view(batch_size, node_num, -1)
        # Multiply the encoded vectors by the cluster assignment to get the final node representations
        # [batch size, cluster_number, hidden dimension]
        node_repr = torch.bmm(assignment.transpose(1, 2), encoded)
        return node_repr, assignment

    def target_distribution(self, batch: torch.Tensor) -> torch.Tensor:
        """
        Compute the target distribution p_ij, given the batch (q_ij), as in 3.1.3 Equation 3 of
        Xie/Girshick/Farhadi; this is used the KL-divergence loss function.

        :param batch: [batch size, number of clusters] Tensor of dtype float
        :return: [batch size, number of clusters] Tensor of dtype float
        """
        weight = (batch**2) / torch.sum(batch, 0)
        return (weight.t() / torch.sum(weight, 1)).t()

    def loss(self, assignment):
        flat_q = assignment.view(-1, assignment.size(-1))       # q
        p = self.target_distribution(flat_q).detach()           # p (no grad)
        return self.loss_fn(flat_q.log(), p)

    def get_cluster_centers(self) -> torch.Tensor:
        """
        Get the cluster centers, as computed by the encoder.

        :return: [number of clusters, hidden dimension] Tensor of dtype float
        """
        return self.assignment.get_cluster_centers()


class ClusterAssignment(nn.Module):
    def __init__(
        self,
        cluster_number: int,
        embedding_dimension: int,
        alpha: float = 1.0,
        cluster_centers = None,
        orthogonal=True,
        freeze_center=True,
        project_assignment=True,
    ) -> None:
        """
        Module to handle the soft assignment, for a description see in 3.1.1. in Xie/Girshick/Farhadi,
        where the Student's t-distribution is used measure similarity between feature vector and each
        cluster centroid.

        :param cluster_number: number of clusters
        :param embedding_dimension: embedding dimension of feature vectors
        :param alpha: parameter representing the degrees of freedom in the t-distribution, default 1.0
        :param cluster_centers: clusters centers to initialise, if None then use Xavier uniform
        """
        super(ClusterAssignment, self).__init__()
        self.embedding_dimension = embedding_dimension
        self.cluster_number = cluster_number
        self.alpha = alpha
        self.project_assignment = project_assignment
        if cluster_centers is None:
            initial_cluster_centers = torch.zeros(
                self.cluster_number, self.embedding_dimension, dtype=torch.float
            )
            nn.init.xavier_uniform_(initial_cluster_centers)

        else:
            initial_cluster_centers = cluster_centers

        if orthogonal:
            orthogonal_cluster_centers = torch.zeros(
                self.cluster_number, self.embedding_dimension, dtype=torch.float
            )
            orthogonal_cluster_centers[0] = initial_cluster_centers[0]
            for i in range(1, cluster_number):
                project = 0
                for j in range(i):
                    project += self.project(
                        initial_cluster_centers[j], initial_cluster_centers[i]
                    )
                initial_cluster_centers[i] -= project
                orthogonal_cluster_centers[i] = initial_cluster_centers[i] / torch.norm(
                    initial_cluster_centers[i], p=2
                )

            initial_cluster_centers = orthogonal_cluster_centers

        self.cluster_centers = Parameter(
            initial_cluster_centers, requires_grad=(not freeze_center)
        )

    @staticmethod
    def project(u, v):
        return (torch.dot(u, v) / torch.dot(u, u)) * u

    def forward(self, batch: torch.Tensor) -> torch.Tensor:
        """
        Compute the soft assignment for a batch of feature vectors, returning a batch of assignments
        for each cluster.

        :param batch: FloatTensor of [batch size, embedding dimension]
        :return: FloatTensor [batch size, number of clusters]
        """

        if self.project_assignment:
            assignment = batch @ self.cluster_centers.T
            # prove
            assignment = torch.pow(assignment, 2)

            norm = torch.norm(self.cluster_centers, p=2, dim=-1)
            soft_assign = assignment / norm
            return softmax(soft_assign, dim=-1)

        else:
            norm_squared = torch.sum(
                (batch.unsqueeze(1) - self.cluster_centers) ** 2, 2
            )
            numerator = 1.0 / (1.0 + (norm_squared / self.alpha))
            power = float(self.alpha + 1) / 2
            numerator = numerator**power
            return numerator / torch.sum(numerator, dim=1, keepdim=True)

    def get_cluster_centers(self) -> torch.Tensor:
        """
        Get the cluster centers.

        :return: FloatTensor [number of clusters, embedding dimension]
        """
        return self.cluster_centers


class InterpretableTransformerEncoder(TransformerEncoderLayer):
    def __init__(
        self,
        d_model,
        nhead,
        dim_feedforward=2048,
        dropout=0.1,
        activation=F.relu,
        layer_norm_eps=1e-5,
        batch_first=False,
        norm_first=False,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__(
            d_model,
            nhead,
            dim_feedforward,
            dropout,
            activation,
            layer_norm_eps,
            batch_first,
            norm_first,
            device,
            dtype,
        )
        self.attention_weights = None

    def _sa_block(
        self,
        x,
        attn_mask,
        key_padding_mask,
        is_causal: bool = False,
    ):
        x, weights = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=True,
        )
        self.attention_weights = weights
        return self.dropout1(x)

    def get_attention_weights(self):
        return self.attention_weights

    


# pylint: disable=line-too-long
""" A script for downloading a few fMRI datasets that were uploaded to github. 
    It will download a variant of ABIDE1 dataset (it is smaller than the one we used in our work)\
    and COBRE dataset, both ready for experiments within this framework.
    To use them in experiments, set `dataset=abide` or `dataset=cobre` when running `scripts/run_experiments.py`. 
"""
import requests
import os

from pathlib import Path
DATA_ROOT = Path(os.path.dirname(os.path.dirname(__file__))).joinpath("assets/data")

items = [
    (
        "https://raw.githubusercontent.com/UsmanMahmood27/MILC/master/Data/ABIDE1_AllData.h5",
        DATA_ROOT.joinpath("abide/ABIDE1_AllData.h5"),
    ),
    (
        "https://raw.githubusercontent.com/UsmanMahmood27/MILC/master/IndicesAndLabels/labels_ABIDE1.csv",
        DATA_ROOT.joinpath("abide/labels_ABIDE1.csv"),
    ),
    (
        "https://raw.githubusercontent.com/UsmanMahmood27/MILC/master/Data/COBRE_AllData.h5",
        DATA_ROOT.joinpath("cobre/COBRE_AllData.h5"),
    ),
    (
        "https://raw.githubusercontent.com/UsmanMahmood27/MILC/master/IndicesAndLabels/labels_COBRE.csv",
        DATA_ROOT.joinpath("cobre/labels_COBRE.csv"),
    ),
]

for item in items:
    r = requests.get(item[0], allow_redirects=True, timeout=100)
    os.makedirs(os.path.dirname(item[1]), exist_ok=True)
    with open(item[1], "wb") as f:
        f.write(r.content)


"""
Collection of helper functions, common model routines, and a training script.
"""

import numpy as np
import pandas as pd

import torch
from torch.nn.functional import cross_entropy, softmax
from torch.utils.data import DataLoader, TensorDataset

from sklearn.metrics import roc_auc_score
from sklearn.metrics import (
    accuracy_score, balanced_accuracy_score, f1_score,
    roc_auc_score,
)

import time
from copy import deepcopy


def basic_ce_loss(logits, targets):
    """
    Default cross entropy loss function used by many models.
    
    Returns
    -------
    loss : Tensor
        CE loss for backprop.
    logs : dict
        Loss dictionary for logs.
    """
    loss = cross_entropy(logits, targets)

    return loss, {"CE_loss": float(loss.detach().cpu().item())}


def basic_Adam_optimizer(model, lr):
    """ Basic Adam optimizer """
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    return optimizer


def compute_metrics(y_prob, y_pred, y_true):
    """
    Compute a bundle of classification metrics.
    """
    log = {}
    log["accuracy"] = accuracy_score(y_true, y_pred)
    log["balanced_accuracy"] = balanced_accuracy_score(y_true, y_pred)
    if y_prob.shape[1] == 2: # binary classification
        log["auc"] = roc_auc_score(y_true, y_prob[:, 1])
    else: # multiclass classification
        log["auc"] = roc_auc_score(y_true, y_prob, multi_class='ovr', average='macro')
    log["f1_macro"] = f1_score(y_true, y_pred, average="macro")

    return log


def basic_handle_batch(model, batch):
    """
    Standard batch handling routine for models. 
    Works for the majority of models that expect single input (time series or FNC).
    Only exception is FBNetGen that expects both time series and FNC.

    Returns
    -------
    loss : Tensor
        Loss for backprop (if you need it)
    batch_log : dict
        Dictionary of classification metrics and losses for logs.
    """
    # load batch into model
    
    data, labels = batch[:-1], batch[-1] # most of the time batch contains (data, labels), but sometimes there's more
    logits, loss_load = model(*data)
    loss, loss_log = model.compute_loss(loss_load, labels)

    # compute metrics
    y_prob = softmax(logits, dim=1).detach().cpu().numpy()
    y_pred = y_prob.argmax(axis=1)
    y_true = labels.detach().cpu().numpy()

    batch_log = compute_metrics(y_prob, y_pred, y_true)
    batch_log.update(loss_log)

    return loss, batch_log


def basic_dataloader(data,
                     labels,
                     type,
                     batch_size: int = 64,
                     shuffle: bool = True,
                     zscore=True):
    """
    Dataloader factory for time series and FNC data.
    Creates a DataLoader from time series data and labels, derives FNC as PCC if needed.
    Most of the models hook their `prepare_dataloader` methods to this function.

    Args
    ----
    data : array-like, shape (B, T, D)
        Time series data of shape Batch x Time x (D)Features 
    labels : array-like, shape (B,)
        Class labels for the data.
    type : str
        Determines the type of data stored in the dataloader:
        "TS" for time series, "FNC" for connectivity matrices, "ALL" for both.
    batch_size : int, optional
        Batch size for the DataLoader.
    shuffle : bool, optional
        Whether to shuffle batching in the DataLoader.
    zscore : bool, optional
        Whether to z-score time series.

    Returns
    -------
    DataLoader
        A PyTorch DataLoader generating the batches of data (as TS or FNC, or both) and labels.
    """

    if type == "TS":
        if zscore:
            data = zscore_np(data, axis=1) # (B, T, D)
        data = torch.tensor(data, dtype=torch.float32)
        labels = torch.tensor(labels, dtype=torch.int64)
        dataset = TensorDataset(data, labels)
    elif type == "FNC":
        fnc = corrcoef_batch(data)  # (B, D, D)
        fnc = torch.tensor(fnc, dtype=torch.float32)
        labels = torch.tensor(labels, dtype=torch.int64)
        dataset = TensorDataset(fnc, labels)
    elif type == "ALL":
        if zscore:
            data = zscore_np(data, axis=1) # (B, T, D)
        fnc = corrcoef_batch(data)  # (B, D, D)
        data = torch.tensor(data, dtype=torch.float32)
        fnc = torch.tensor(fnc, dtype=torch.float32)
        labels = torch.tensor(labels, dtype=torch.int64)
        dataset = TensorDataset(data, fnc, labels)
    else:
        raise ValueError(f"Unknown data type: {type}. Supported types are 'TS', 'FNC', and 'ALL'.")
        
    return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)


def zscore_np(x, axis=0, eps=1e-8):
    """
    numpy-only z-score

    Parameters
    ----------
    x : array_like
        data to z-score
    axis : int or tuple of int
        Axis/axes along which to compute mean/std. (e.g., axis=1 for (B,T,D) time axis)

    Returns
    -------
    xz : ndarray
        Z-scored array (same shape as x).
    """
    x = np.asarray(x)
    mu = np.mean(x, axis=axis, keepdims=True)
    sd = np.std(x, axis=axis, keepdims=True)
    return (x - mu) / (sd + eps)


def corrcoef_batch(x, eps=1e-8):
    """
    Batched Pearson correlation over features for time-series data.

    Parameters
    ----------
    X : array_like, shape (B, T, D)
        Data to correlate, arranged as Batch x Time x (D)Features.

    Returns
    -------
    C : ndarray, shape (B, D, D)
        Pearson correlation matrices per batch item.
    """
    x = np.asarray(x)
    B, T, D = x.shape

    # z-score along time
    Z = zscore_np(x, axis=1, eps=eps)   # (B, T, D)

    # Correlation = (Z^T @ Z) / T for each batch element
    C = np.einsum('btd,bte->bde', Z, Z) / T             # (B, D, D)

    # Symmetrize & fix diagonals (nice-to-have for numeric drift)
    C = 0.5 * (C + C.transpose(0, 2, 1))
    batch_indices = np.arange(B)[:, None]
    diag_indices = np.arange(D)
    C[batch_indices, diag_indices, diag_indices] = 1.0

    return C


class BasicTrainer:
    """
    Basic training loop using model-provided helpers. 
    Trains the model for a given number of epochs but stops if validation loss doesn't improve for `patience` epochs.

    Args
    ----
    model : torch.nn.Module
        One of the models from `ml4fmri.models` package.
    train_loader, val_loader, test_loader : DataLoader
        Dataloaders that yield batches with data appropriate for the model. `model.prepare_dataloader` can be used to create them.
    epochs : int, optional
        Max epochs (default: 200).
    lr : float, optional
        Optimizer LR (default: use model's LR).
    device : str, optional
        "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
    patience : int, optional
        Early stopping patience on validation loss (default: 30).

    Returns on `run()`
    ------------------
    train_df : pandas.DataFrame
        Training log useful for training analysis. Can be used to check training curves.
    test_df : pandas.DataFrame
        Test log with classification scores.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        train_loader: DataLoader,
        val_loader: DataLoader,
        test_loader: DataLoader,
        epochs: int = 200,
        lr: float = None,
        device: str = None,
        patience: int = 30,
    ) -> None:
        
        self.model = model
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.test_loader = test_loader

        self.epochs = epochs
        if lr is None:
            lr = model.lr
        self.lr = lr
        if device is not None:
            self.device = torch.device(device)
        else:
            self.device = torch.device("cuda") if torch.cuda.is_available() \
                else torch.device("mps") if torch.backends.mps.is_available() \
                    else torch.device("cpu")
        self.patience = patience


        # Optimizer
        self.optimizer = model.get_optimizer(lr)

        # bookkeeping
        self.n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        self.best_state = None
        self.best_val_loss = None

        # move model to device
        self.model.to(self.device)


    def _epoch(self, loader: DataLoader, train: bool):
        with torch.set_grad_enabled(train):
            self.model.train(train)
            total_loss = 0.0
            n_batches = len(loader)
            agg_log = {}

            for batch in loader:
                batch = tuple(t.to(self.device) for t in batch) # cast batch to device
                loss, batch_log = self.model.handle_batch(batch)

                total_loss += float(loss.detach().cpu().item())

                if train:
                    self.optimizer.zero_grad()
                    loss.backward()
                    self.optimizer.step()

                for k, v in batch_log.items():
                    if isinstance(v, (int, float)):
                        agg_log[k] = agg_log.get(k, 0.0) + float(v)

            # average logs over batches
            agg_log = {k: v / n_batches for k, v in agg_log.items()}
            agg_log["loss"] = total_loss / n_batches
            return agg_log

    def run(self):
        train_logs = []
        stopper = EarlyStopping(patience=self.patience) # monitors training, stops if validation loss does not improve

        ### Training loop
        start = time.time()
        for epoch in range(self.epochs):
            train_log = self._epoch(self.train_loader, train=True)
            val_log = self._epoch(self.val_loader, train=False)

            # handle logs
            epoch_log = {"model": self.model.__class__.__name__, "epoch": epoch, "lr": self.lr}
            epoch_log.update({f"train_{k}": v for k, v in train_log.items()})
            epoch_log.update({f"val_{k}": v for k, v in val_log.items()})
            train_logs.append(epoch_log)

            # track training via validation loss
            improved = stopper.step(val_log["loss"])
            if improved:
                self.best_state = deepcopy(self.model.state_dict())


            if stopper.stop:
                break
        training_time = time.time() - start
        
        # build train history DataFrame
        train_logs = pd.DataFrame(train_logs)

        # restore best weights
        if self.best_state is not None:
            self.model.load_state_dict(self.best_state)


        ### Test loop
        test_logs = {"model": self.model.__class__.__name__}
        test_log = self._epoch(self.test_loader, train=False)
        test_log = {f"test_{k}": v for k, v in test_log.items()}
        test_logs.update(test_log)
        test_logs.update({
            "train_time": training_time,
            "n_params": self.n_params,
        })
        test_logs = pd.DataFrame([test_logs])

        return train_logs, test_logs

class EarlyStopping:
    """Early stopping mechanism, watches if metric is minimized."""
    def __init__(self, patience: int = 30):
        self.patience = patience
        self.best_score = None
        self.counter = 0
        self.stop = False

    def step(self, value: float) -> bool:
        """
        Returns a boolean indicating if the metric improved.
        Toggles stop flag if patience is exceeded.
        """
        if self.best_score is None or value < self.best_score - 1e-12:
            self.best_score = value
            self.counter = 0
            return True  # improved
        self.counter += 1
        if self.counter >= self.patience:
            self.stop = True
        return False


# tests/test_models_api.py
import importlib
import pandas as pd
import torch

def _first_batch(dataloader):
    for b in dataloader:
        return b
    return None

def test_exported_models_api(toy_data, dims, toy_splits):
    X, y = toy_data
    D, C = dims
    idx = toy_splits

    mdl = importlib.import_module("ml4fmri.models")
    names = list(getattr(mdl, "__all__", []))
    assert names, "ml4fmri.models.__all__ is empty; export your model classes there"

    # Models that don't follow standard PyTorch API (e.g., sklearn-based)
    excluded_models = ["LR"]

    failures = []

    for name in names:
        Model = getattr(mdl, name)

        def fail(msg):
            failures.append(f"[{name}] {msg}")

        # Skip excluded models entirely
        if name in excluded_models:
            continue

        # 1) constructor: (input_size, output_size)
        try:
            m = Model(D, C)
            if not isinstance(m, torch.nn.Module):
                fail("constructor did not produce a torch.nn.Module")
        except Exception as e:
            fail(f"constructor raised: {e}")
            continue  # skip the rest for this model

        # 2) dataloaders via class helper 
        if not hasattr(Model, "prepare_dataloader"):
            fail("missing class method prepare_dataloader")
            continue
        try:
            Xtr, ytr = X[idx["train"]], y[idx["train"]]
            Xva, yva = X[idx["val"]],   y[idx["val"]]
            Xte, yte = X[idx["test"]],  y[idx["test"]]

            train_dl = Model.prepare_dataloader(Xtr, ytr, shuffle=True)
            val_dl   = Model.prepare_dataloader(Xva, yva, shuffle=False)
            test_dl  = Model.prepare_dataloader(Xte, yte, shuffle=False)
        except Exception as e:
            fail(f"prepare_dataloader raised: {e}")
            continue

        # 3) one batch
        try:
            batch = _first_batch(train_dl)
            if batch is None:
                fail("train dataloader yielded no batches")
                continue

            bx, by = batch[:-1], batch[-1] # most of the time batch contains (data, labels), but sometimes there's more
            if not isinstance(by, torch.Tensor):
                fail("batch elements are not torch.Tensors")
        except Exception as e:
            fail(f"consuming first batch raised: {e}")
            continue

        # 4) forward returns (logits, dict)
        try:
            logits, loss_load = m(*bx)
            if not (isinstance(logits, torch.Tensor) and logits.ndim == 2 and logits.shape[1] == C):
                fail(f"forward: logits has wrong shape {tuple(getattr(logits,'shape',()))}; expected (*, {C})")
            if not isinstance(loss_load, dict):
                fail("forward: second return is not a dict (loss_load)")
        except Exception as e:
            fail(f"forward raised: {e}")
            continue

        # 5) compute_loss works
        if not hasattr(m, "compute_loss"):
            fail("missing compute_loss")
        else:
            try:
                loss, loss_log = m.compute_loss(loss_load, by)
                if not (isinstance(loss, torch.Tensor) and torch.isfinite(loss).item()):
                    fail("compute_loss: loss is not finite")
                if not isinstance(loss_log, dict):
                    fail("compute_loss: logs is not a dict")
            except Exception as e:
                fail(f"compute_loss raised: {e}")

        # 6) handle_batch works
        if not hasattr(m, "handle_batch"):
            fail("missing handle_batch")
        else:
            try:
                loss2, blog = m.handle_batch(batch)
                if not (isinstance(loss2, torch.Tensor) and torch.isfinite(loss2).item()):
                    fail("handle_batch: loss is not finite")
                if not isinstance(blog, dict):
                    fail("handle_batch: blog is not a dict")
            except Exception as e:
                fail(f"handle_batch raised: {e}")

        # 7) get_optimizer (optional)
        if hasattr(m, "get_optimizer"):
            try:
                opt = m.get_optimizer(lr=getattr(m, "lr", 1e-3))
                import torch.optim as optim
                if not isinstance(opt, optim.Optimizer):
                    fail("get_optimizer did not return a torch.optim.Optimizer")
            except Exception as e:
                fail(f"get_optimizer raised: {e}")

        # 8) train_model (tiny run)
        if not hasattr(m, "train_model"):
            fail("missing train_model")
        else:
            try:
                tdf, sdf = m.train_model(train_dl, val_dl, test_dl, epochs=2, patience=1)
                if not (isinstance(tdf, pd.DataFrame) and "epoch" in tdf.columns):
                    fail("train_model: train_df is not a DataFrame with 'epoch'")
                if not isinstance(sdf, pd.DataFrame):
                    fail("train_model: test_df is not a DataFrame")
            except Exception as e:
                fail(f"train_model raised: {e}")

    # final summary
    assert not failures, "Model API failures:\n  - " + "\n  - ".join(failures)



# pylint: disable=invalid-name, missing-function-docstring, missing-class-docstring, unused-argument, too-few-public-methods, no-member, too-many-arguments, line-too-long, too-many-instance-attributes, too-many-locals
"""
FBNetGen model module

Works simultaneously with time series and FNC data.

@inproceedings{fbnetgen,
    title={{FBNETGEN}: Task-aware {GNN}-based f{MRI} Analysis via Functional Brain Network Generation},
    author={Xuan Kan and Hejie Cui and Joshua Lukemire and Ying Guo and Carl Yang},
    booktitle={Medical Imaging with Deep Learning},
    year={2022},
    url={https://openreview.net/forum?id=oWFphg2IKon}
}
"""

from types import SimpleNamespace
import numpy as np

from torch.nn.functional import softmax, cross_entropy
from torch import nn, optim
from torch.nn import Conv1d, MaxPool1d, Linear, GRU
from torch import nn
import torch

from .helper_functions import basic_handle_batch, basic_dataloader, BasicTrainer

class FBNetGen(nn.Module):
    """
    TIME SERIES + FNC model
    FBNetGen model for fMRI data from https://doi.org/10.48550/arXiv.2205.12465
    Original code: https://github.com/Wayfear/BrainNetworkTransformer

    Expected input shape: [batch_size, time_length, input_feature_size] for time series, [batch_size, input_feature_size, input_feature_size] for FNC.
    Output: [batch_size, n_classes], loss_load = {"logits": logits}
    """
    def __init__(self,
                 input_size,
                 output_size,
        ):
        """
        Initialize FBNetGen model.
        Args:
            input_size (int): Size of the vector at each time step in the input time series. \
                Common to all models, for FNC models it is the number of nodes (width or length) in the FNC matrix.
            output_size (int): Number of classes for classification. \
                Common to all models.
        """

        super().__init__()

        self.lr = 1e-4 # learning rate used in the paper. Defined like this in every model for reference.
        self.weight_decay = 1e-4 # weight decay used by the optimizer

        # FBNetGen loss configuration
        group_loss = True
        sparsity_loss = True
        sparsity_loss_weight = 0.1
        self.criterion = FBNetGenLoss(group_loss, sparsity_loss, sparsity_loss_weight)

        model_cfg = {
            "extractor_type": "gru",
            "embedding_size": 16,
            "window_size": 4,
            "cnn_pool_size": 16,
            "graph_generation": "product",  # product or linear
            "num_gru_layers": 4,
            "dropout": 0.5,
            # loss
            "group_loss": True,
            "sparsity_loss": True,
            "sparsity_loss_weight": 1e-4,
            # data shape
            "timeseries_sz": None, # some FBNetGen modules need time length to initialize (extractor_type = 'cnn')
            "node_sz": input_size,
            "node_feature_sz": input_size,
            "output_size": output_size,
            # optimizer
            "optimizer": {"lr": self.lr, "weight_decay": self.weight_decay},
        }

        def dict_to_namespace(d):
            """Convert nested dict to nested SimpleNamespace"""
            if isinstance(d, dict):
                return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
            return d
        
        model_cfg = dict_to_namespace(model_cfg)

        assert model_cfg.extractor_type in ["cnn", "gru"]
        assert model_cfg.graph_generation in ["linear", "product"]

        self.graph_generation = model_cfg.graph_generation
        if model_cfg.extractor_type == "cnn":
            self.extract = ConvKRegion(
                out_size=model_cfg.embedding_size,
                kernel_size=model_cfg.window_size,
                time_series=model_cfg.timeseries_sz,
            )
        elif model_cfg.extractor_type == "gru":
            self.extract = GruKRegion(
                out_size=model_cfg.embedding_size,
                kernel_size=model_cfg.window_size,
                layers=model_cfg.num_gru_layers,
            )
        if self.graph_generation == "linear":
            self.emb2graph = Embed2GraphByLinear(
                model_cfg.embedding_size, roi_num=model_cfg.node_sz
            )
        elif self.graph_generation == "product":
            self.emb2graph = Embed2GraphByProduct(
                model_cfg.embedding_size, roi_num=model_cfg.node_sz
            )

        self.predictor = GNNPredictor(
            model_cfg.node_feature_sz,
            roi_num=model_cfg.node_sz,
            n_classes=model_cfg.output_size,
        )

    def forward(self, time_series, node_feature):
        """
        Args:
            time_series: fMRI time series, shape (B, T, D)
            node_feature: FNC matrices derived from time series, shape (B, D, D)
        """
        time_series = time_series.transpose(1, 2)

        x = self.extract(time_series)
        x = softmax(x, dim=-1)
        m = self.emb2graph(x)

        logits = self.predictor(m, node_feature)
        loss_load = {
            "logits": logits,
            "learned_matrix": m
        }

        return logits, loss_load

    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """

        loss, loss_log = self.criterion(loss_load, targets)
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """
        
        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        optimizer = optim.Adam(
            self.parameters(),
            lr=lr,
            weight_decay=self.weight_decay,
        )

        return optimizer

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True,
                           window_size = 4):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        # FBNetGen needs time series length to be divisible by window_size hyperparam
        T = data.shape[1]
        if T % window_size != 0:
            rem = T % window_size
            data = data[:, :-rem, :]

        return basic_dataloader(data,
                                labels,
                                type="ALL",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """

        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs
    

class FBNetGenLoss:
    def __init__(self, group_loss=True, sparsity_loss=True, sparsity_loss_weight=0.1):
        self.group_loss = group_loss
        self.sparsity_loss = sparsity_loss
        self.sparsity_loss_weight = sparsity_loss_weight

    def __call__(self, loss_load, target):
        logits = loss_load["logits"]
        learned_matrix = loss_load["learned_matrix"]

        ce_loss = cross_entropy(logits, target)
        loss_log = {"ce_loss": ce_loss.detach().cpu().item()}
        loss = 2 * ce_loss

        if self.group_loss:
            group_loss = 2 * self.intra_loss(target, learned_matrix) + self.inner_loss(target, learned_matrix)
            loss_log["group_loss"] = group_loss.detach().cpu().item()
            loss += group_loss

        if self.sparsity_loss:
            sparsity_loss = learned_matrix.abs().mean()
            loss_log["sparsity_loss"] = sparsity_loss.detach().cpu().item()
            loss += self.sparsity_loss_weight * sparsity_loss

        return loss, loss_log

    def inner_loss(self, label, matrices):
        loss = 0

        if torch.sum(label == 0) > 1:
            loss += torch.mean(torch.var(matrices[label == 0], dim=0))

        if torch.sum(label == 1) > 1:
            loss += torch.mean(torch.var(matrices[label == 1], dim=0))

        return loss

    def intra_loss(self, label, matrices):
        a, b = None, None

        if torch.sum(label == 0) > 0:
            a = torch.mean(matrices[label == 0], dim=0)

        if torch.sum(label == 1) > 0:
            b = torch.mean(matrices[label == 1], dim=0)

        if a is not None and b is not None:
            return 1 - torch.mean(torch.pow(a - b, 2))

        return 0
    

class GruKRegion(nn.Module):
    def __init__(self, kernel_size=8, layers=4, out_size=8, dropout=0.5):
        super().__init__()
        self.gru = GRU(
            kernel_size, kernel_size, layers, bidirectional=True, batch_first=True
        )

        self.kernel_size = kernel_size

        self.linear = nn.Sequential(
            nn.Dropout(dropout),
            Linear(kernel_size * 2, kernel_size),
            nn.LeakyReLU(negative_slope=0.2),
            Linear(kernel_size, out_size),
        )

    def forward(self, raw):
        b, k, _ = raw.shape

        x = raw.reshape(b * k, -1, self.kernel_size)

        x, _ = self.gru(x)

        x = x[:, -1, :]

        x = x.view((b, k, -1))

        x = self.linear(x)

        return x


class ConvKRegion(nn.Module):
    def __init__(self, k=1, out_size=8, kernel_size=8, pool_size=16, time_series=512):
        super().__init__()
        self.conv1 = Conv1d(
            in_channels=k, out_channels=32, kernel_size=kernel_size, stride=2
        )

        output_dim_1 = (time_series - kernel_size) // 2 + 1

        self.conv2 = Conv1d(in_channels=32, out_channels=32, kernel_size=8)
        output_dim_2 = output_dim_1 - 8 + 1
        self.conv3 = Conv1d(in_channels=32, out_channels=16, kernel_size=8)
        output_dim_3 = output_dim_2 - 8 + 1
        self.max_pool1 = MaxPool1d(pool_size)
        output_dim_4 = output_dim_3 // pool_size * 16
        self.in0 = nn.InstanceNorm1d(time_series)
        self.in1 = nn.BatchNorm1d(32)
        self.in2 = nn.BatchNorm1d(32)
        self.in3 = nn.BatchNorm1d(16)

        self.linear = nn.Sequential(
            Linear(output_dim_4, 32),
            nn.LeakyReLU(negative_slope=0.2),
            Linear(32, out_size),
        )

    def forward(self, x):
        b, k, d = x.shape

        x = torch.transpose(x, 1, 2)

        x = self.in0(x)

        x = torch.transpose(x, 1, 2)
        x = x.contiguous()

        x = x.view((b * k, 1, d))

        x = self.conv1(x)

        x = self.in1(x)
        x = self.conv2(x)

        x = self.in2(x)
        x = self.conv3(x)

        x = self.in3(x)
        x = self.max_pool1(x)

        x = x.view((b, k, -1))

        x = self.linear(x)

        return x


class Embed2GraphByProduct(nn.Module):
    def __init__(self, input_dim, roi_num=264):
        super().__init__()

    def forward(self, x):
        return x @ x.transpose(1, 2)


class GNNPredictor(nn.Module):
    def __init__(self, node_input_dim, roi_num=360, n_classes=2):
        super().__init__()
        inner_dim = roi_num
        self.roi_num = roi_num
        self.gcn = nn.Sequential(
            nn.Linear(node_input_dim, inner_dim),
            nn.LeakyReLU(negative_slope=0.2),
            Linear(inner_dim, inner_dim),
        )
        self.bn1 = torch.nn.BatchNorm1d(inner_dim)

        self.gcn1 = nn.Sequential(
            nn.Linear(inner_dim, inner_dim),
            nn.LeakyReLU(negative_slope=0.2),
        )
        self.bn2 = torch.nn.BatchNorm1d(inner_dim)
        self.gcn2 = nn.Sequential(
            nn.Linear(inner_dim, 64),
            nn.LeakyReLU(negative_slope=0.2),
            nn.Linear(64, 8),
            nn.LeakyReLU(negative_slope=0.2),
        )
        self.bn3 = torch.nn.BatchNorm1d(inner_dim)

        self.fcn = nn.Sequential(
            nn.Linear(8 * roi_num, 256),
            nn.LeakyReLU(negative_slope=0.2),
            nn.Linear(256, 32),
            nn.LeakyReLU(negative_slope=0.2),
            nn.Linear(32, n_classes),
        )

    def forward(self, m, node_feature):
        bz = m.shape[0]

        x = torch.einsum("ijk,ijp->ijp", m, node_feature)

        x = self.gcn(x)

        x = x.reshape((bz * self.roi_num, -1))
        x = self.bn1(x)
        x = x.reshape((bz, self.roi_num, -1))

        x = torch.einsum("ijk,ijp->ijp", m, x)

        x = self.gcn1(x)

        x = x.reshape((bz * self.roi_num, -1))
        x = self.bn2(x)
        x = x.reshape((bz, self.roi_num, -1))

        x = torch.einsum("ijk,ijp->ijp", m, x)

        x = self.gcn2(x)

        x = self.bn3(x)

        x = x.view(bz, -1)

        return self.fcn(x)
    


class Embed2GraphByLinear(nn.Module):
    def __init__(self, input_dim, roi_num=360):
        super().__init__()

        self.fc_out = nn.Linear(input_dim * 2, input_dim)
        self.fc_cat = nn.Linear(input_dim, 1)

        def encode_onehot(labels):
            classes = set(labels)
            classes_dict = {
                c: np.identity(len(classes))[i, :] for i, c in enumerate(classes)
            }
            labels_onehot = np.array(
                list(map(classes_dict.get, labels)), dtype=np.int32
            )
            return labels_onehot

        off_diag = np.ones([roi_num, roi_num])
        rel_rec = np.array(encode_onehot(np.where(off_diag)[0]), dtype=np.float32)
        rel_send = np.array(encode_onehot(np.where(off_diag)[1]), dtype=np.float32)
        self.rel_rec = torch.FloatTensor(rel_rec).cuda()
        self.rel_send = torch.FloatTensor(rel_send).cuda()

    def forward(self, x):
        batch_sz, region_num, _ = x.shape
        receivers = torch.matmul(self.rel_rec, x)

        senders = torch.matmul(self.rel_send, x)
        x = torch.cat([senders, receivers], dim=2)
        x = torch.relu(self.fc_out(x))
        x = self.fc_cat(x)

        x = torch.relu(x)

        m = torch.reshape(x, (batch_sz, region_num, region_num, -1))
        return m



# pylint: disable=invalid-name, missing-function-docstring
""" 
MILC model module.

CNN+LSTM model for fMRI data, can work with any time series.

@inproceedings{wholeMILC,
    author = {Mahmood, Usman and Rahman, Md Mahfuzur and Fedorov, Alex and Lewis, Noah and Fu, Zening and Calhoun, Vince D. and Plis, Sergey M.},
    title = {Whole MILC: Generalizing Learned Dynamics Across Tasks, Datasets, and Populations},
    year = {2020},
    isbn = {978-3-030-59727-6},
    publisher = {Springer-Verlag},
    address = {Berlin, Heidelberg},
    url = {https://doi.org/10.1007/978-3-030-59728-3_40},
    doi = {10.1007/978-3-030-59728-3_40},
    booktitle = {Medical Image Computing and Computer Assisted Intervention – MICCAI 2020: 23rd International Conference, Lima, Peru, October 4–8, 2020, Proceedings, Part VII},
    pages = {407–417},
    numpages = {11},
    keywords = {Resting state fMRI., Deep learning, Self-supervised, Transfer learning},
    location = {Lima, Peru}
}
"""

from sys import path
import torch
from torch import nn
from torch.nn.functional import cross_entropy, softmax
from types import SimpleNamespace
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from .helper_functions import basic_handle_batch, basic_dataloader, BasicTrainer, zscore_np
    
class MILCLoss:
    """Cross-entropy loss with model regularization"""

    def __init__(self, reg_param: float = 1e-6):

        self.reg_param = reg_param

    def __call__(self, loss_load, targets):
        logits = loss_load["logits"]
        model = loss_load["model"]
        device = logits.device

        ce_loss = cross_entropy(logits, targets)

        # L1 reg
        reg_loss = torch.zeros((), device=device)
        for module in (model.lstm, model.attn):
            for name, p in module.named_parameters():
                if "bias" not in name:
                    reg_loss = reg_loss + p.abs().sum()

        loss = ce_loss + self.reg_param * reg_loss
        return loss, {
            "ce_loss": float(ce_loss.detach().cpu().item()),
            "reg_loss": float(reg_loss.detach().cpu().item()),
        }
    
class MILC(nn.Module):
    """
    TIME SERIES MODEL
    MILC model for fMRI data from https://doi.org/10.48550/arXiv.2007.16041.
    Orig implementation: https://github.com/UsmanMahmood27/MILC.
    Expected input shape: [batch_size, n_windows, feature_size, window_size] (see prepare_dataloader for details)
    Output: [batch_size, n_classes]
    """

    def __init__(self,
                 input_size: int,
                 output_size: int
        ):
        super().__init__()

        self.lr = 2e-4  # learning rate used in the paper. Defined like this in every model for reference.
        self.eps = 1e-5 # epsilon for Adam optimizer
        self.reg_param = 1e-5  # regularization parameter, used in loss
        self.criterion = MILCLoss(self.reg_param)

        model_cfg = {
            "data_params": {
                "input_size": input_size,
                "output_size": output_size,
                "window_size": 20,
                "window_shift": 10,
            },
            "lstm": {
                "input_feature_size": 256, 
                "hidden_size": 200, 
                "n_layers": 1
            },
            "reg_param": 1e-5,
        }
        def dict_to_namespace(d):
            """Convert nested dict to nested SimpleNamespace"""
            if isinstance(d, dict):
                return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
            return d
        
        model_cfg = dict_to_namespace(model_cfg)


        self.encoder = NatureOneCNN(model_cfg)
        self.lstm = LSTM(model_cfg)
        self.attn = nn.Sequential(
            nn.Linear(2 * model_cfg.lstm.hidden_size, 128),
            nn.ReLU(),
            nn.Linear(128, 1),
        )

        self.decoder = nn.Sequential(
            nn.Linear(model_cfg.lstm.hidden_size, model_cfg.lstm.hidden_size),
            nn.ReLU(),
            nn.Linear(model_cfg.lstm.hidden_size, model_cfg.data_params.output_size),
        )

        self.init_weight()

    def init_weight(self):
        for name, param in self.decoder.named_parameters():
            if "weight" in name:
                nn.init.xavier_normal_(param, gain=0.65)
        for name, param in self.attn.named_parameters():
            if "weight" in name:
                nn.init.xavier_normal_(param, gain=0.65)

    def get_attention(self, x):
        B, N, H = x.shape

        # last window embedding per sample: (B, 1, H) so it can broadcast across N
        x_last = x[:, -1:, :]                       # (B, 1, H)

        # Concatenate [x_i, x_last] for every i in a vectorized way: (B, N, 2H)
        pairs = torch.cat([x, x_last.expand(-1, N, -1)], dim=-1)

        # MLP over last dim (nn.Linear handles (..., in_features))
        weights = self.attn(pairs).squeeze(-1)      # (B, N)

        # Normalize across windows
        weights = softmax(weights, dim=1)           # (B, N)

        # Weighted sum across windows -> (B, H)
        # (B,N,H) * (B,N,1) -> sum over N
        attended = (x * weights.unsqueeze(-1)).sum(dim=1)
        return attended

    def forward(self, x):
        bs, nw, fs, ws = x.shape  # [batch_size, n_windows, feature_size, window_size]

        encoder_output = self.encoder(x.view(-1, fs, ws))
        encoder_output = encoder_output.view(bs, nw, -1)

        lstm_output = self.lstm(encoder_output)

        attention_output = self.get_attention(lstm_output)

        logits = self.decoder(attention_output)
    
        loss_load = {
            "logits": logits,
            "model": self,
        }
        return logits, loss_load

    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """
        loss, loss_log = self.criterion(loss_load, targets)

        return loss, loss_log
    
    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        optimizer = torch.optim.Adam(self.parameters(), lr=lr, eps=self.eps)
        return optimizer

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True,
                           window_size: int = 20,
                           window_shift: int = 10,
                           zscore: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to handle_batch)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        if zscore:
            data = zscore_np(data, axis=1)

        # transform data into overlapping windows of shape (Batch, (tau)Windows, Feature_size, (W)Window_size) 
        win_data = sliding_window_view(data, window_shape=window_size, axis=1)  # (B, Tau, D, W)
        win_data = win_data[:, ::window_shift, :, :]                            # (B, tau, D, W)

        return basic_dataloader(win_data,
                                labels,
                                type="TS",
                                batch_size=batch_size,
                                shuffle=shuffle,
                                zscore=False)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """

        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs
    

class LSTM(nn.Module):
    """Bidirectional LSTM for classifying subjects."""

    def __init__(self, model_cfg):
        super().__init__()
        input_size = model_cfg.lstm.input_feature_size
        self.hidden_size = model_cfg.lstm.hidden_size
        n_layers = model_cfg.lstm.n_layers

        self.lstm = nn.LSTM(
            input_size,
            self.hidden_size // 2,
            num_layers=n_layers,
            bidirectional=True,
            batch_first=True,
        )

        self.init_weight()

    def init_weight(self):
        for name, param in self.lstm.named_parameters():
            if "weight" in name:
                nn.init.xavier_normal_(param, gain=0.65)  # 0.65 is default gain

    def forward(self, x):
        lstm_output, _ = self.lstm(x)
        return lstm_output
    
class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.shape[0], -1)


class NatureOneCNN(nn.Module):
    def __init__(self, model_cfg):
        super().__init__()

        # original feature size is treated as a number of channels of a 1D image
        feature_size = model_cfg.data_params.input_size
        # 1D image has size of the window time length
        window_size = model_cfg.data_params.window_size
        # encoder output is passed to LSTM
        output_size = model_cfg.lstm.input_feature_size

        # calculate CNN layers output
        cnn_output_size = window_size - 3
        cnn_output_size = cnn_output_size - 3
        cnn_output_size = cnn_output_size - 2
        final_conv_size = 200 * cnn_output_size

        self.cnn = nn.Sequential(
            self.init_module(nn.Conv1d(feature_size, 64, 4, stride=1)),
            nn.ReLU(),
            self.init_module(nn.Conv1d(64, 128, 4, stride=1)),
            nn.ReLU(),
            self.init_module(nn.Conv1d(128, 200, 3, stride=1)),
            nn.ReLU(),
            Flatten(),
            self.init_module(nn.Linear(final_conv_size, output_size)),
        )

    def init_module(self, module):
        # weight_init
        nn.init.orthogonal_(module.weight.data, gain=nn.init.calculate_gain("relu"))
        # bias_init
        nn.init.constant_(module.bias.data, 0)

        return module

    def forward(self, inputs):
        return self.cnn(inputs)


# pylint: disable=invalid-name, missing-function-docstring
""" 
LSTM model module.

LSTM-based classifier, applicable to any time series.
Fairly generic design, this one was used in:

@article{meanMLP,
    title = {A simple but tough-to-beat baseline for fMRI time-series classification},
    journal = {NeuroImage},
    volume = {303},
    pages = {120909},
    year = {2024},
    issn = {1053-8119},
    doi = {10.1016/j.neuroimage.2024.120909},
    url = {https://www.sciencedirect.com/science/article/pii/S1053811924004063},
    author = {Pavel Popov and Usman Mahmood and Zening Fu and Carl Yang and Vince Calhoun and Sergey Plis},
}

More original paper (Schmidhuber!!):

@article{lstm,
    author = {Hochreiter, Sepp and Schmidhuber, Jürgen},
    title = {Long Short-Term Memory},
    journal = {Neural Computation},
    volume = {9},
    number = {8},
    pages = {1735-1780},
    year = {1997},
    month = {11},
    issn = {0899-7667},
    doi = {10.1162/neco.1997.9.8.1735},
    url = {https://doi.org/10.1162/neco.1997.9.8.1735},
    eprint = {https://direct.mit.edu/neco/article-pdf/9/8/1735/813796/neco.1997.9.8.1735.pdf},
}
"""

import torch
from torch import nn

from .helper_functions import basic_ce_loss, basic_handle_batch, basic_dataloader, basic_Adam_optimizer, BasicTrainer

class LSTM(nn.Module):
    """
    TIME SERIES MODEL

    Vanilla-style LSTM classifier for fMRI data from https://doi.org/10.1016/j.neuroimage.2024.120909.
    Expected input shape: [batch_size, time_length, input_feature_size].
    Output: [batch_size, n_classes]
    """

    def __init__(
            self,
            input_size: int,
            output_size: int,
            hidden_size: int = 210,
            num_layers: int = 1,
            bidirectional: bool = True,
            dropout: float = 0.5,
            lr: float = 4e-5,
    ):
        """
        Initialize LSTM model.
        Args:
            input_size (int): Size of the vector at each time step in the input time series. \
                Common to all models, for FNC models it is the number of nodes (width or length) in the FNC matrix.
            output_size (int): Number of classes for classification. \
                Common to all models.
            hidden_size (int, hyperparameter): LSTM hidden size. Defaults to 210.
            num_layers (int, hyperparameter): Number of LSTM layers. Defaults to 1.
            bidirectional (bool, hyperparameter): Whether to use a bidirectional LSTM. Defaults to True.
            dropout (float, hyperparameter): Dropout rate for classifier. Defaults to 0.5.
            lr (float, hyperparameter): Learning rate for the optimizer. Defaults to 4e-5. \
                Common to all models. Isn't used by the model per se, presented as a reference LR from the paper. 
        """
        super().__init__()
        self.lr = lr  # learning rate used in the paper. Defined like this in every model for reference.

        self.bidirectional = bidirectional
        self.hidden_size = hidden_size

        self.lstm = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            batch_first=True,
            bidirectional=bidirectional,
        )
        self.fc = nn.Sequential(
            nn.Dropout(p=dropout),
            nn.Linear(
                2 * hidden_size if bidirectional else hidden_size,
                output_size,
            ),
        )

    def forward(self, x):
        lstm_output, _ = self.lstm(x)

        if self.bidirectional:
            out_forward = lstm_output[:, -1, : self.hidden_size]
            out_reverse = lstm_output[:, 0, self.hidden_size :]
            lstm_output = torch.cat((out_forward, out_reverse), 1)
        else:
            lstm_output = lstm_output[:, -1, :]

        logits = self.fc(lstm_output)

        return logits, {"logits": logits}
    
    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """

        loss, loss_log = basic_ce_loss(loss_load["logits"], targets)
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        return basic_Adam_optimizer(self, lr)

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        return basic_dataloader(data,
                                labels,
                                type="TS",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """
        
        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )

        train_logs, test_logs = trainer.run()
        return train_logs, test_logs



# pylint: disable=invalid-name, missing-function-docstring, missing-class-docstring, unused-argument, too-few-public-methods, no-member, too-many-arguments, line-too-long, too-many-instance-attributes
"""
BrainNetCNN model module

Works with FNC data.

@article{brainnetcnn,
    title = {BrainNetCNN: Convolutional neural networks for brain networks; towards predicting neurodevelopment},
    journal = {NeuroImage},
    volume = {146},
    pages = {1038-1049},
    year = {2017},
    issn = {1053-8119},
    doi = {10.1016/j.neuroimage.2016.09.046},
    url = {https://www.sciencedirect.com/science/article/pii/S1053811916305237},
    author = {Jeremy Kawahara and Colin J. Brown and Steven P. Miller and Brian G. Booth and Vann Chau and Ruth E. Grunau and Jill G. Zwicker and Ghassan Hamarneh},
}
"""

from torch.nn import functional as F
from torch import nn, optim
import torch

from .helper_functions import basic_ce_loss, basic_handle_batch, basic_dataloader, BasicTrainer

class BrainNetCNN(nn.Module):
    """
    FNC MODEL

    BrainNetCNN model from https://doi.org/10.1016/j.neuroimage.2016.09.046.
    Orig implementation: https://github.com/Wayfear/BrainNetworkTransformer.
    Expected input shape: [batch_size, input_feature_size, input_feature_size].
    Output: [batch_size, n_classes]
    """
    def __init__(self,
                 input_size: int,
                 output_size: int
        ):
        super().__init__()
        self.lr = 1e-4  # learning rate used in the paper. Defined like this in every model for reference.
        self.weight_decay = 1e-4  # weight decay used by optimizer
    
        self.in_planes = 1
        self.d = input_size

        self.e2econv1 = E2EBlock(1, 32, self.d, bias=True)
        self.e2econv2 = E2EBlock(32, 64, self.d, bias=True)
        self.E2N = torch.nn.Conv2d(64, 1, (1, self.d))
        self.N2G = torch.nn.Conv2d(1, 256, (self.d, 1))
        self.dense1 = torch.nn.Linear(256, 128)
        self.dense2 = torch.nn.Linear(128, 30)
        self.dense3 = torch.nn.Linear(30, output_size)

    def forward(self, node_feature: torch.tensor):
        node_feature = node_feature.unsqueeze(dim=1)
        out = F.leaky_relu(self.e2econv1(node_feature), negative_slope=0.33)
        out = F.leaky_relu(self.e2econv2(out), negative_slope=0.33)
        out = F.leaky_relu(self.E2N(out), negative_slope=0.33)
        out = F.dropout(F.leaky_relu(self.N2G(out), negative_slope=0.33), p=0.5)
        out = out.view(out.size(0), -1)
        out = F.dropout(F.leaky_relu(self.dense1(out), negative_slope=0.33), p=0.5)
        out = F.dropout(F.leaky_relu(self.dense2(out), negative_slope=0.33), p=0.5)
        logits = F.leaky_relu(self.dense3(out), negative_slope=0.33)

        return logits, {"logits": logits}

    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """

        loss, loss_log = basic_ce_loss(loss_load["logits"], targets)
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
            
        optimizer = optim.Adam(
            self.parameters(),
            lr=lr,
            weight_decay=self.weight_decay,
        )

        return optimizer


    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        return basic_dataloader(data,
                                labels,
                                type="FNC",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """

        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs
    

class E2EBlock(torch.nn.Module):
    """E2Eblock."""

    def __init__(self, in_planes, planes, roi_num, bias=True):
        super().__init__()
        self.d = roi_num
        self.cnn1 = torch.nn.Conv2d(in_planes, planes, (1, self.d), bias=bias)
        self.cnn2 = torch.nn.Conv2d(in_planes, planes, (self.d, 1), bias=bias)

    def forward(self, x):
        a = self.cnn1(x)
        b = self.cnn2(x)
        return torch.cat([a] * self.d, 3) + torch.cat([b] * self.d, 2)



import importlib
import inspect
import pkgutil
import torch.nn as nn

def _is_model_class(obj):
    return (
        inspect.isclass(obj)
        and (issubclass(obj, nn.Module) or obj.__name__ == "LR")
        and hasattr(obj, "prepare_dataloader")
        and hasattr(obj, "train_model")
    )

def test_models_all_is_complete():
    # import/refresh the package
    mdl = importlib.import_module("ml4fmri.models")
    importlib.invalidate_caches()
    importlib.reload(mdl)

    # Collect exported names
    exported = set(getattr(mdl, "__all__", []))

    # Discover implemented classes by scanning submodules
    discovered = {}
    if hasattr(mdl, "__path__"):
        for info in pkgutil.iter_modules(mdl.__path__):
            modname = info.name
            if modname.startswith("_"):
                continue
            try:
                mod = importlib.import_module(f"{mdl.__name__}.{modname}")
            except Exception:
                continue

            # STRICT: class with same name as module
            cls = getattr(mod, modname, None)
            if _is_model_class(cls):
                discovered[cls.__name__] = cls
                continue

    # Assert every discovered model is exported in __all__
    assert set(discovered.keys()).issubset(exported), (
        "These implemented model classes are not exported in ml4fmri.models.__all__: "
        f"{sorted(set(discovered.keys()) - exported)}"
    )

    # Assert each exported name resolves to the actual class object
    for name in exported:
        obj = getattr(mdl, name, None)
        assert _is_model_class(obj), f"Exported name '{name}' is not a valid model class"

        # If discovered, ensure identity matches (same object)
        if name in discovered:
            assert obj is discovered[name], (
                f"Exported '{name}' does not match class found in its module"
            )



# pylint: disable=invalid-name, missing-function-docstring, no-member
""" 
meanTransformer model module.

Transformer-encoder based classifier with encoder output averaging, applicable to any time series.
Was used in:

@article{meanMLP,
    title = {A simple but tough-to-beat baseline for fMRI time-series classification},
    journal = {NeuroImage},
    volume = {303},
    pages = {120909},
    year = {2024},
    issn = {1053-8119},
    doi = {10.1016/j.neuroimage.2024.120909},
    url = {https://www.sciencedirect.com/science/article/pii/S1053811924004063},
    author = {Pavel Popov and Usman Mahmood and Zening Fu and Carl Yang and Vince Calhoun and Sergey Plis},
}
"""
import math
import torch
from torch import nn

from .helper_functions import basic_ce_loss, basic_handle_batch, basic_dataloader, basic_Adam_optimizer, BasicTrainer


class meanTransformer(nn.Module):
    """
    TIME SERIES MODEL
    meanTransformer model for fMRI data from https://doi.org/10.1016/j.neuroimage.2024.120909.
    Like Transformer model, but with temporal averaging of output embeddings.
    Expected input shape: [batch_size, time_length, input_feature_size].
    Output: [batch_size, n_classes]
    """
    
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 head_hidden_size: int = 20,
                 num_heads: int = 4,
                 num_layers: int = 2,
                 dropout: float = 0.3,
                 lr: float = 0.002
    ):
        """
        Initialize meanTransformer model.
        Args:
            input_size (int): Size of the vector at each time step in the input time series.
            output_size (int): Number of classes for classification.
            head_hidden_size (int, hyperparameter): Size of each attention head. Defaults to 20.
            num_heads (int, hyperparameter): Number of attention heads. Defaults to 3.
            num_layers (int, hyperparameter): Number of transformer encoder layers. Defaults to 2.
            dropout (float, hyperparameter): Dropout rate for classifier. Defaults to 0.3.
            lr (float, hyperparameter): Learning rate for the optimizer. Defaults to 0.002.
                Common to all models. Isn't used by the model per se, presented as a reference LR from the paper.
        """

        super().__init__()
        self.lr = lr  # learning rate used in the paper. Defined like this in every model for reference.

        hidden_size = head_hidden_size * num_heads

        encoder_layer = nn.TransformerEncoderLayer(
            d_model=hidden_size, nhead=num_heads, batch_first=True
        )
        self.transformer = nn.TransformerEncoder(encoder_layer, num_layers=num_layers)

        self.input_embed = nn.Linear(input_size, hidden_size)
        self.pos_encoder = PositionalEncoding(hidden_size)

        self.clf = nn.Sequential(
            nn.Dropout(p=dropout), nn.Linear(hidden_size, output_size)
        )

    def forward(self, x):
        input_embed = self.input_embed(x)
        input_embed = self.pos_encoder(input_embed)

        tf_output = self.transformer(input_embed)
        tf_output = tf_output.mean(dim=1)  # temporal averaging

        logits = self.clf(tf_output)

        return logits, {"logits": logits}


    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """

        loss, loss_log = basic_ce_loss(loss_load["logits"], targets)
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        return basic_Adam_optimizer(self, lr)

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        return basic_dataloader(data,
                                labels,
                                type="TS",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """
        
        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )

        train_logs, test_logs = trainer.run()
        return train_logs, test_logs

class PositionalEncoding(nn.Module):
    """Positional encoding module"""

    def __init__(self, input_feature_size, max_seq_length=2048):
        super().__init__()

        # Create a matrix of shape (max_seq_length, input_feature_size)
        pe = torch.zeros(max_seq_length, input_feature_size)

        # Calculate the positional encoding
        position = torch.arange(0, max_seq_length, dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, input_feature_size, 2, dtype=torch.float32)
            * -(math.log(10000.0) / input_feature_size)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(1).transpose(
            0, 1
        )  # Transpose for (batch_size, max_seq_length, d_model)
        self.register_buffer("pe", pe)

    def forward(self, x):
        # expects data of size [batch_size, time_length, input_feature_size]
        return x + self.pe[:, : x.size(1), :]



# pylint: disable=invalid-name, missing-function-docstring
""" 
meanMLP model module.

Simple time series model, surprisingly good on fMRI classification.

This is also a showcase module that explains the general API for all models.
It is commented in greater detail, presents and explains the calls common to all models.

Each module docstring contains .bib item of the paper that presented the respective model.
For meanMLP, it is:

@article{meanMLP,
    title = {A simple but tough-to-beat baseline for fMRI time-series classification},
    journal = {NeuroImage},
    volume = {303},
    pages = {120909},
    year = {2024},
    issn = {1053-8119},
    doi = {10.1016/j.neuroimage.2024.120909},
    url = {https://www.sciencedirect.com/science/article/pii/S1053811924004063},
    author = {Pavel Popov and Usman Mahmood and Zening Fu and Carl Yang and Vince Calhoun and Sergey Plis},
}
"""

import torch
from torch import nn

from .helper_functions import basic_ce_loss, basic_handle_batch, basic_dataloader, basic_Adam_optimizer, BasicTrainer

class meanMLP(nn.Module):
    """
    TIME SERIES MODEL

    meanMLP model for fMRI data from https://doi.org/10.1016/j.neuroimage.2024.120909
    
    Expected input shape: [batch_size, time_length, input_feature_size].
    Output: [batch_size, n_classes], loss_load = {"logits": logits}
    """

    def __init__(
            self,
            input_size: int,
            output_size: int,
            hidden_size: int = 160,
            dropout: float = 0.49,
            lr: float = 0.005,
    ):
        """
        Initialize meanMLP model.
        Args:
            input_size (int): Size of the vector at each time step in the input time series. \
                Common to all models, for FNC models it is the number of nodes (width or length) in the FNC matrix.
            output_size (int): Number of classes for classification. \
                Common to all models.
            hidden_size (int, hyperparameter): Size of the middle MLP layer. Defaults to 160.
            dropout (float, hyperparameter): Dropout rate. Defaults to 0.49.
            lr (float, hyperparameter): Learning rate for the optimizer. Defaults to 0.005. \
                Common to all models. Isn't used by the model per se, presented as a reference LR from the paper. 
        """

        super().__init__()
        self.lr = lr # learning rate used in the paper. Defined like this in every model for reference.

        self.mlp = nn.Sequential(
            nn.Linear(input_size, hidden_size),
            nn.LayerNorm(hidden_size),
            nn.ReLU(),
            nn.Dropout(p=dropout),
            nn.Linear(hidden_size, output_size),
        )

    def forward(self, x: torch.Tensor):
        # bs, tl, fs = x.shape  # [batch_size, time_length, input_feature_size]

        ts_logits = self.mlp(x)
        logits = ts_logits.mean(1)

        return logits, {"logits": logits} # all models return (1) logits and (2) a dictionary of data for computing loss

    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """

        loss, loss_log = basic_ce_loss(loss_load["logits"], targets)
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        return basic_Adam_optimizer(self, lr)

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        return basic_dataloader(data,
                                labels,
                                type="TS",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """

        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs


# pylint: disable=invalid-name, missing-function-docstring
""" 
DICE model module.

LSTM-based connectivity estimator and classifier, works with time series.

@article{dice,
    title = {Through the looking glass: Deep interpretable dynamic directed connectivity in resting {fMRI}},
    journal = {NeuroImage},
    volume = {264},
    pages = {119737},
    year = {2022},
    issn = {1053-8119},
    doi = {10.1016/j.neuroimage.2022.119737},
    url = {https://www.sciencedirect.com/science/article/pii/S1053811922008588},
    author = {Usman Mahmood and Zening Fu and Satrajit Ghosh and Vince Calhoun and Sergey Plis},
}
"""

import torch
from torch import nn
from torch.nn.functional import cross_entropy
from types import SimpleNamespace

from .helper_functions import basic_handle_batch, basic_dataloader, basic_Adam_optimizer, BasicTrainer

class DICELoss:
    """Cross-entropy loss with model regularization"""

    def __init__(self, reg_param: float = 1e-6):

        self.reg_param = reg_param

    def __call__(self, loss_load, targets):
        logits = loss_load["logits"]
        model = loss_load["model"]
        device = logits.device

        ce_loss = cross_entropy(logits, targets)

        # L1 reg
        reg_loss = torch.zeros((), device=device)
        for module in (model.gta_embed, model.gta_attend):
            for name, p in module.named_parameters():
                if "bias" not in name:
                    reg_loss = reg_loss + p.abs().sum()

        loss = ce_loss + self.reg_param * reg_loss
        return loss, {
            "ce_loss": float(ce_loss.detach().cpu().item()),
            "reg_loss": float(reg_loss.detach().cpu().item()),
        }

class DICE(nn.Module):
    """
    TIME SERIES MODEL
    DICE model for fMRI data from https://doi.org/10.1016/j.neuroimage.2022.119737.
    Orig implementation: https://github.com/UsmanMahmood27/DICE.
    Expected input shape: [batch_size, time_length, input_feature_size].
    Output: [batch_size, n_classes]
    """

    def __init__(self,
                 input_size: int,
                 output_size: int
        ):
        super().__init__()

        self.lr = 2e-4  # learning rate used in the paper. Defined like this in every model for reference.
        self.reg_param = 1e-6  # regularization parameter, used in loss
        self.criterion = DICELoss(self.reg_param)

        model_cfg = {
            "lstm": {
                "bidirectional": True,
                "num_layers": 1,
                "hidden_size": 50,
            },
            "clf": {
                "hidden_size": 64,
                "num_layers": 0,
            },
            "MHAtt": {
                "n_heads": 1,
                "head_hidden_size": 48,
                "dropout": 0.0,
            },
            "scheduler": {
                "patience": 4,
                "factor": 0.5,
            },
            "reg_param": 1e-6,
            "lr": 2e-4,
            "input_size": input_size,
            "output_size": output_size,
        }
        
        def dict_to_namespace(d):
            """Convert nested dict to nested SimpleNamespace"""
            if isinstance(d, dict):
                return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
            return d
        
        model_cfg = dict_to_namespace(model_cfg)

        lstm_hidden_size = model_cfg.lstm.hidden_size
        lstm_num_layers = model_cfg.lstm.num_layers
        bidirectional = model_cfg.lstm.bidirectional

        self.lstm_output_size = (
            lstm_hidden_size * 2 if bidirectional else lstm_hidden_size
        )

        clf_hidden_size = model_cfg.clf.hidden_size
        clf_num_layers = model_cfg.clf.num_layers

        MHAtt_n_heads = model_cfg.MHAtt.n_heads
        MHAtt_hidden_size = MHAtt_n_heads * model_cfg.MHAtt.head_hidden_size
        MHAtt_dropout = model_cfg.MHAtt.dropout

        # LSTM - first block
        self.lstm = nn.LSTM(
            input_size=1,
            hidden_size=lstm_hidden_size,
            num_layers=lstm_num_layers,
            bidirectional=bidirectional,
            batch_first=True,
        )

        # Classifier - last block
        clf = [
            nn.Linear(input_size**2, clf_hidden_size),
            nn.ReLU(),
        ]
        for _ in range(clf_num_layers):
            clf.append(nn.Linear(clf_hidden_size, clf_hidden_size))
            clf.append(nn.ReLU())
        clf.append(
            nn.Linear(clf_hidden_size, output_size),
        )
        self.clf = nn.Sequential(*clf)

        # Multihead attention - second block
        self.key_layer = nn.Sequential(
            nn.Linear(
                self.lstm_output_size,
                MHAtt_hidden_size,
            ),
        )
        self.value_layer = nn.Sequential(
            nn.Linear(
                self.lstm_output_size,
                MHAtt_hidden_size,
            ),
        )
        self.query_layer = nn.Sequential(
            nn.Linear(
                self.lstm_output_size,
                MHAtt_hidden_size,
            ),
        )
        self.multihead_attn = nn.MultiheadAttention(
            embed_dim=MHAtt_hidden_size,
            num_heads=MHAtt_n_heads,
            dropout=MHAtt_dropout,
            batch_first=True,
        )

        # Global Temporal Attention - third block
        self.upscale = 0.05
        self.upscale2 = 0.5

        self.HW = torch.nn.Hardswish()
        self.gta_embed = nn.Sequential(
            nn.Linear(
                input_size**2,
                round(self.upscale * input_size**2),
            ),
        )
        self.gta_norm = nn.Sequential(
            nn.BatchNorm1d(round(self.upscale * input_size**2)),
            nn.ReLU(),
        )
        self.gta_attend = nn.Sequential(
            nn.Linear(
                round(self.upscale * input_size**2),
                round(self.upscale2 * input_size**2),
            ),
            nn.ReLU(),
            nn.Linear(round(self.upscale2 * input_size**2), 1),
        )

        self.init_weight()

    def init_weight(self):
        for module in (
            self.lstm,
            self.clf,
            self.query_layer,
            self.key_layer,
            self.value_layer,
            self.multihead_attn,
        ):
            for name, param in module.named_parameters():
                if "weight" in name:
                    nn.init.kaiming_normal_(param, mode="fan_in")

        for module in (
            self.gta_embed,
            self.gta_attend,
        ):
            for name, param in module.named_parameters():
                if "weight" in name and param.dim() > 1:
                    nn.init.kaiming_normal_(param, mode="fan_in")
       
    def gta_attention(self, x, node_axis=1):
        # x.shape: [batch_size; time_length; input_feature_size * input_feature_size]
        x_readout = x.mean(node_axis, keepdim=True)
        x_readout = x * x_readout

        a = x_readout.shape[0]
        b = x_readout.shape[1]
        x_readout = x_readout.reshape(-1, x_readout.shape[2])
        x_embed = self.gta_norm(self.gta_embed(x_readout))
        x_graphattention = (self.gta_attend(x_embed).squeeze()).reshape(a, b)
        x_graphattention = self.HW(x_graphattention.reshape(a, b))
        return (x * (x_graphattention.unsqueeze(-1))).mean(node_axis)

    def multi_head_attention(self, x):
        # x.shape: [time_length * batch_size; input_feature_size; lstm_hidden_size]
        key = self.key_layer(x)
        value = self.value_layer(x)
        query = self.query_layer(x)

        attn_output, attn_output_weights = self.multihead_attn(key, value, query)

        return attn_output, attn_output_weights

    def forward(self, x):
        B, T, C = x.shape  # [batch_size, time_length, input_feature_size]

        # 1. pass input to LSTM; treat each channel as an independent single-feature time series
        x = x.permute(0, 2, 1)  # x.shape: [batch_size; input_feature_size; time_length]
        x = x.reshape(B * C, T, 1)  # x.shape: [batch_size * n_channels; time_length; 1]
        ##########################
        lstm_output, _ = self.lstm(x)
        # lstm_output.shape: [batch_size * input_feature_size; time_length; lstm_hidden_size]
        ##########################
        lstm_output = lstm_output.reshape(B, C, T, self.lstm_output_size)
        # lstm_output.shape: [batch_size; input_feature_size; time_length; lstm_hidden_size]

        # 2. pass lstm_output at each time point to multihead attention to reveal spatial connctions
        lstm_output = lstm_output.permute(2, 0, 1, 3)
        # lstm_output.shape: [time_length; batch_size; input_feature_size; lstm_hidden_size]
        lstm_output = lstm_output.reshape(T * B, C, self.lstm_output_size)
        # lstm_output.shape: [time_length * batch_size; input_feature_size; lstm_hidden_size]
        ##########################
        _, attn_weights = self.multi_head_attention(lstm_output)
        # attn_weights.shape: [time_length * batch_size; input_feature_size; input_feature_size]
        ##########################
        attn_weights = attn_weights.reshape(T, B, C, C)
        # attn_weights.shape: [time_length; batch_size; input_feature_size; input_feature_size]
        attn_weights = attn_weights.permute(1, 0, 2, 3)
        # attn_weights.shape: [batch_size; time_length; input_feature_size; input_feature_size]

        # 3. pass attention weights to a global temporal attention to obrain global graph
        attn_weights = attn_weights.reshape(B, T, -1)
        # attn_weights.shape: [batch_size; time_length; input_feature_size * input_feature_size]
        ##########################
        FC = self.gta_attention(attn_weights)
        # FC.shape: [batch_size; input_feature_size * input_feature_size]
        ##########################

        # 4. Pass learned graph to the classifier to get predictions
        logits = self.clf(FC)
        # logits.shape: [batch_size; n_classes]

        loss_load = {
            "logits": logits,
            "model": self,
        }
        return logits, loss_load

    #### Helper functions for model training and evaluation ####

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.
        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.
        
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """
        loss, loss_log = self.criterion(loss_load, targets)

        return loss, loss_log
    
    def handle_batch(self, batch):
        """
        Standard batch handling routine for models. 
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.
        
        Args:
            batch (tuple): A batch containing the time series data and labels as a tuple.
        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """

        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        return basic_Adam_optimizer(self, lr)

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`)
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            DataLoader: A PyTorch DataLoader generating the batches of time series data and labels. 
        """
        return basic_dataloader(data,
                                labels,
                                type="TS",
                                batch_size=batch_size,
                                shuffle=shuffle)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs: int = 200,
              lr: float = None,
              device: str = None,
              patience: int = 30,
        ):
        """
        Standard model training routine.
        Args:
            train_loader (DataLoader): DataLoader for the training set. Used for training.
            val_loader (DataLoader): DataLoader for the validation set. Used during training to find most generalizable model.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device to train the model on: "cuda", "mps", or "cpu". Default: auto-detect (cuda -> mps -> cpu).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """

        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs


# pylint: disable=invalid-name, missing-function-docstring
""" 
Logistic regression model module.

Classic logistic regression model, wired for FNC classification.

@article{logit,
    author = {Joseph Berkson},
    title = {Application of the Logistic Function to Bio-Assay},
    journal = {Journal of the American Statistical Association},
    volume = {39},
    number = {227},
    pages = {357--365},
    year = {1944},
    doi = {10.1080/01621459.1944.10500699},
    URL = {https://doi.org/10.1080/01621459.1944.10500699},
    eprint = {https://doi.org/10.1080/01621459.1944.10500699}
}
"""

import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
import time

from .helper_functions import corrcoef_batch, compute_metrics

class LR():
    """
    FNC MODEL

    Basic logistic regression model for fMRI data.
    Based on the sklearn's logistic regression, basically a wrapper that adds cvbench-required API.
    Expected input shape: [batch_size, input_feature_size, input_feature_size].
    Output: [batch_size, n_classes], loss_load = {"logits": logits}
    """

    def __init__(
            self,
            input_size: int,
            output_size: int,
    ):
        """
        Initialize LR model.
        """
        self.model = LogisticRegression()
    def __call__(self, x):
        return self.forward(x)
    
    def forward(self, x):
        """ For compatibility """
        return 0, {"logits": None} # all models return (1) logits and (2) a dictionary of data for computing loss

    def compute_loss(self, loss_load, targets):
        """ For compatibility """

        loss, loss_log = 0, {"dummy_loss": 0}
        return loss, loss_log

    def handle_batch(self, batch):
        """ For compatibility """

        loss, batch_log = 0, {"dummy_loss": 0, "dummy_score": 1}
        return loss, batch_log
    
    def get_optimizer(self, lr=None):
        """ For compatibility """
        return None

    @staticmethod
    def prepare_dataloader(data,
                           labels,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        Returns FNC and label numpy arrays that can be used for training of sklearn's LR.
        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.
        
        Returns:
            (fnc, labels): FNC and labels as numpy arrays.
        """
        fnc = corrcoef_batch(data)
        n_rois = fnc.shape[1]
        tril_idx = np.tril_indices(n_rois, k=-1)
        fnc = fnc[:, tril_idx[0], tril_idx[1]]

        return (fnc, labels)
    
    def train_model(self,
              train_loader,
              val_loader,
              test_loader,
              epochs= None,
              lr= None,
              device= None,
              patience= None,
        ):
        """
        Standard model training routine, rewired for LR.
        Args:
            train_loader (fnc, labels): Tuple of numpy arrays for LR training returned by prepare_dataloader
            val_loader (fnc, labels): Merged with train data
            test_loader (fnc, labels): Test data
            epochs (optional): Kept for compatibility
            lr (optional): Kept for compatibility
            device (optional): Kept for compatibility
            patience (optional): Kept for compatibility
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """
        
        train_data, train_labels = train_loader
        val_data, val_labels = val_loader
        test_data, test_labels = test_loader

        # merge training and validation sets
        train_data = np.concatenate([train_data, val_data], axis=0)
        train_labels = np.concatenate([train_labels, val_labels], axis=0)

        # train
        start_time = time.time()
        self.model.fit(
            X=train_data,
            y=train_labels,
        )
        training_time = time.time() - start_time

        # train log
        train_logs = {"model": "LR", "epoch": 0, "lr": 0}
        y_score = self.model.predict_proba(train_data)
        y_pred = np.argmax(y_score, axis=-1).astype(np.int32)
        train_log = compute_metrics(y_true=train_labels, y_pred=y_pred, y_prob=y_score)
        train_logs.update({f"train_{k}": v for k, v in train_log.items()})
        train_logs.update({f"val_{k}": v for k, v in train_log.items()})
        train_loss = np_cross_entropy(y_score, train_labels)
        train_logs.update({"train_loss": train_loss, "val_loss": train_loss})
        train_logs = pd.DataFrame([train_logs])
        
        # test
        y_score = self.model.predict_proba(test_data)
        y_pred = np.argmax(y_score, axis=-1).astype(np.int32)
        test_logs = {"model": "LR"}
        test_log = compute_metrics(y_true=test_labels, y_pred=y_pred, y_prob=y_score)
        test_logs.update({f"test_{k}": v for k, v in test_log.items()})
        test_logs.update({
            "train_time": training_time,
            "n_params": self.model.coef_.size,
        })
        test_loss = np_cross_entropy(y_score, test_labels)
        test_logs.update({"test_loss": test_loss})
        test_logs = pd.DataFrame([test_logs])

        return train_logs, test_logs


import numpy as np

def np_cross_entropy(y_prob, y_true, eps: float = 1e-12):
    B, C = y_prob.shape

    p_true = y_prob[np.arange(B), y_true]
    loss = -np.log(np.clip(p_true, eps, 1.0))

    return float(loss.mean())



# pylint: disable=invalid-name, missing-function-docstring
"""
BolT model module.

Sliding window transformer model for fMRI time series analysis. 
Fairly good, one of the few models that outperformed meanMLP on large datasets.

@article{bolT,
    title = {BolT: Fused window transformers for fMRI time series analysis},
    journal = {Medical Image Analysis},
    volume = {88},
    pages = {102841},
    year = {2023},
    issn = {1361-8415},
    doi = {10.1016/j.media.2023.102841},
    url = {https://www.sciencedirect.com/science/article/pii/S1361841523001019},
    author = {Hasan A. Bedel and Irmak Sivgin and Onat Dalmaz and Salman U.H. Dar and Tolga Çukur},
}

"""

import math
from types import SimpleNamespace

import torch
from torch import nn
from torch.nn.functional import cross_entropy
from torch.nn.init import trunc_normal_

from .helper_functions import (
    basic_handle_batch,
    basic_dataloader,
    basic_Adam_optimizer,
    BasicTrainer,
)


class BolT(nn.Module):
    """
    TIME SERIES MODEL
    BolT model for fMRI data from https://doi.org/10.1016/j.media.2023.102841.
    Original: https://github.com/icon-lab/BolT
    This version focuses on classification only (analysis/interpretability code removed).

    Expected input shape: [batch_size, time_length, input_feature_size].
    Output: logits [batch_size, n_classes], loss_load = {"logits": logits, "cls": cls}.
    """

    def __init__(self, input_size, output_size):
        """
        Initialize BolT model.
        Majority of hyperparameters are defined in the hyperParams variable.

        Args:
            input_size (int): Size of the vector at each time step in the input time series.
            output_size (int): Number of classes for classification.
        """
        super().__init__()
        self.lr = 2e-4 # learning rate used in the paper. Defined like this in every model for reference.
        self.lambdaCons = 1  # used in loss calculation

        hyperParams = {
            # oneCycleLR hyperparameters from the original code (not used here but kept for reference)
            "lr": 2e-4,
            "minLr": 2e-5,
            "maxLr": 4e-4,
            # BolT core
            "nOfLayers": 4,
            "dim": input_size,
            "numHeads": 36,
            "headDim": 20,
            "windowSize": 20,
            "shiftCoeff": 2.0 / 5.0,
            "fringeCoeff": 2,
            "focalRule": "expand",
            "mlpRatio": 1.0,
            "attentionBias": True,
            "drop": 0.1,
            "attnDrop": 0.1,
            # Loss param
            "lambdaCons": 1,
        }
        hyperParams = SimpleNamespace(**hyperParams)
        self.hyperParams = hyperParams

        dim = input_size
        n_classes = output_size

        self.clsToken = nn.Parameter(torch.zeros(1, 1, dim))

        shiftSize = int(hyperParams.windowSize * hyperParams.shiftCoeff)
        self.shiftSize = shiftSize

        # Build the stack of window-attention blocks
        blocks = []
        for i in range(hyperParams.nOfLayers):
            if hyperParams.focalRule == "expand":
                receptiveSize = hyperParams.windowSize + math.ceil(
                    hyperParams.windowSize * 2 * i * hyperParams.fringeCoeff * (1 - hyperParams.shiftCoeff)
                )
            else:  # "fixed"
                receptiveSize = hyperParams.windowSize + math.ceil(
                    hyperParams.windowSize * 2 * 1 * hyperParams.fringeCoeff * (1 - hyperParams.shiftCoeff)
                )

            blocks.append(
                BolTransformerBlock(
                    dim=dim,
                    numHeads=hyperParams.numHeads,
                    headDim=hyperParams.headDim,
                    windowSize=hyperParams.windowSize,
                    receptiveSize=receptiveSize,
                    shiftSize=shiftSize,
                    mlpRatio=hyperParams.mlpRatio,
                    attentionBias=hyperParams.attentionBias,
                    drop=hyperParams.drop,
                    attnDrop=hyperParams.attnDrop,
                )
            )
        self.blocks = nn.ModuleList(blocks)

        self.encoder_postNorm = nn.LayerNorm(dim)
        self.classifierHead = nn.Linear(dim, n_classes)

        self._initialize_weights()

    def _initialize_weights(self):
        # Initialize CLS token (simple normal, as in the original)
        torch.nn.init.normal_(self.clsToken, std=1.0)

    def forward(self, x):
        # x: [B, T, C]
        B, T, _ = x.shape
        nW = (T - self.hyperParams.windowSize) // self.shiftSize + 1
        cls = self.clsToken.expand(B, nW, -1)

        for block in self.blocks:
            x, cls = block(x, cls)

        cls = self.encoder_postNorm(cls)
        logits = self.classifierHead(cls.mean(dim=1))
        return logits, {"logits": logits, "cls": cls}

    # ==== Helper functions for model training and evaluation ====

    def compute_loss(self, loss_load, targets):
        """
        Standard loss computation routine for models.

        Args:
            loss_load (dict): Forward's second output for the batch.
            targets (torch.Tensor): True labels for the batch.

        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        logs : dict
            Dictionary containing the loss components for logs.
        """
        ce_loss = cross_entropy(loss_load["logits"], targets)

        cls = loss_load["cls"]
        clsLoss = torch.mean(torch.square(cls - cls.mean(dim=1, keepdims=True)))

        loss = ce_loss + clsLoss * self.lambdaCons
        loss_log = {
            "CE_loss": float(ce_loss.detach().cpu().item()),
            "cls_loss": float(clsLoss.detach().cpu().item()),
        }
        return loss, loss_log

    def handle_batch(self, batch):
        """
        Standard batch handling routine for models.
        Returns loss for backprop and a dictionary of classification metrics and losses for logs.

        Args:
            batch (tuple): A batch containing the time series data and labels.

        Returns
        -------
        loss : Tensor
            Loss for backpropagation.
        batch_log : dict
            Dictionary of classification metrics and losses for logs.
        """
        loss, batch_log = basic_handle_batch(self, batch)
        return loss, batch_log

    def get_optimizer(self, lr=None):
        """
        Standard optimizer getter routine for models.
        """
        if lr is None:
            lr = self.lr
        return basic_Adam_optimizer(self, lr)

    @staticmethod
    def prepare_dataloader(data, labels, batch_size: int = 64, shuffle: bool = True):
        """
        Returns torch DataLoader that produces appropriate batches for the model (can pass to `handle_batch`).

        Args:
            data (array-like): Time series data of shape (B, T, D).
            labels (array-like): Class labels for the data.
            batch_size (int, optional): Batch size for the DataLoader. Defaults to 64.
            shuffle (bool, optional): Whether to shuffle batching in the DataLoader. Defaults to True.

        Returns
        -------
            DataLoader
        """
        return basic_dataloader(
            data,
            labels,
            type="TS",
            batch_size=batch_size,
            shuffle=shuffle,
        )

    def train_model(
        self,
        train_loader,
        val_loader,
        test_loader,
        epochs: int = 200,
        lr: float = None,
        device: str = None,
        patience: int = 30,
    ):
        """
        Standard model training routine.

        Args:
            train_loader (DataLoader): DataLoader for the training set.
            val_loader (DataLoader): DataLoader for the validation set.
            test_loader (DataLoader): DataLoader for the test set.
            epochs (int, optional): Number of training epochs. Defaults to 200.
            lr (float, optional): Optimizer learning rate (default: use model's self.lr).
            device (str, optional): Device: "cuda", "mps", or "cpu" (auto-detect if None).
            patience (int, optional): Early stopping patience (in epochs). Defaults to 30.
        Returns
        -------
        (train_logs, test_logs)
            Training and test dataframes containing loss and accuracy metrics.
        """
        trainer = BasicTrainer(
            model=self,
            train_loader=train_loader,
            val_loader=val_loader,
            test_loader=test_loader,
            epochs=epochs,
            lr=lr,
            device=device,
            patience=patience,
        )
        train_logs, test_logs = trainer.run()
        return train_logs, test_logs


# ==== BolT submodules and utilities ====

def windowBoldSignal(boldSignal, windowLength, stride):
    """
    Vectorized windowing via Tensor.unfold.

    boldSignal: (B, N, T)
    return:
        windowed: (B, nW, N, windowLength)
        samplingEndPoints: list[int]  # end indices for each window
    """
    x = boldSignal.unfold(dimension=2, size=windowLength, step=stride)  # (B, N, nW, W)
    x = x.permute(0, 2, 1, 3).contiguous()  # (B, nW, N, W)
    B, nW, N, W = x.shape
    samplingEndPoints = list(range(W, W + nW * stride, stride))
    return x, samplingEndPoints


class FeedForward(nn.Module):
    def __init__(self, dim, mult=1, dropout=0.0):
        super().__init__()
        inner_dim = int(dim * mult)
        activation = nn.GELU()
        self.net = nn.Sequential(
            nn.Linear(dim, inner_dim),
            activation,
            nn.Dropout(dropout),
            nn.Linear(inner_dim, dim),
        )

    def forward(self, x):
        return self.net(x)


def max_neg_value(tensor):
    return -torch.finfo(tensor.dtype).max


class WindowAttention(nn.Module):
    def __init__(
        self,
        dim,
        windowSize,
        receptiveSize,
        numHeads,
        headDim=20,
        attentionBias=True,
        qkvBias=True,
        attnDrop=0.0,
        projDrop=0.0,
    ):
        super().__init__()
        self.dim = dim
        self.windowSize = windowSize  # N
        self.receptiveSize = receptiveSize  # M
        self.numHeads = numHeads
        head_dim = headDim
        self.scale = head_dim ** -0.5
        self.attentionBias = attentionBias

        # relative position bias
        maxDisparity = windowSize - 1 + (receptiveSize - windowSize) // 2
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros(2 * maxDisparity + 1, numHeads)
        )  # (2*maxDisp+1, nH)

        self.cls_bias_sequence_up = nn.Parameter(torch.zeros((1, numHeads, 1, receptiveSize)))
        self.cls_bias_sequence_down = nn.Parameter(torch.zeros(1, numHeads, windowSize, 1))
        self.cls_bias_self = nn.Parameter(torch.zeros((1, numHeads, 1, 1)))

        # precompute pairwise relative position indices for window tokens
        coords_x = torch.arange(self.windowSize)  # N
        coords_x_ = torch.arange(self.receptiveSize) - (self.receptiveSize - self.windowSize) // 2  # M
        relative_coords = coords_x[:, None] - coords_x_[None, :]  # (N, M)
        relative_coords[:, :] += maxDisparity  # shift to start from 0
        relative_position_index = relative_coords  # (N, M)
        self.register_buffer("relative_position_index", relative_position_index)

        self.q = nn.Linear(dim, head_dim * numHeads, bias=qkvBias)
        self.kv = nn.Linear(dim, 2 * head_dim * numHeads, bias=qkvBias)

        self.attnDrop = nn.Dropout(attnDrop)
        self.proj = nn.Linear(head_dim * numHeads, dim)
        self.projDrop = nn.Dropout(projDrop)

        # init learned biases
        trunc_normal_(self.relative_position_bias_table, std=0.02)
        trunc_normal_(self.cls_bias_sequence_up, std=0.02)
        trunc_normal_(self.cls_bias_sequence_down, std=0.02)
        trunc_normal_(self.cls_bias_self, std=0.02)

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, x_, mask, nW):
        """
        x  : (B*nW, 1+windowSize, C)     # queries
        x_ : (B*nW, 1+receptiveSize, C)  # keys/values
        mask: (mask_left, mask_right), each (maskCount, 1+windowSize, 1+receptiveSize), bool
        nW : number of windows

        returns: (B*nW, 1+windowSize, C)
        """
        # shapes / unpack
        BnW, n_tok, C = x.shape  # n_tok = 1 + N
        _, m_tok, _ = x_.shape   # m_tok = 1 + M
        N = n_tok - 1
        M = m_tok - 1
        H = self.numHeads
        B = BnW // nW

        # projections & split heads
        q = self.q(x)  # (B*nW, n_tok, H*d)
        k, v = self.kv(x_).chunk(2, dim=-1)
        d = (self.q.out_features // H)  # per-head dim
        q = q.view(BnW, n_tok, H, d).permute(0, 2, 1, 3).contiguous()   # (B*nW, H, n_tok, d)
        k = k.view(BnW, m_tok, H, d).permute(0, 2, 1, 3).contiguous()   # (B*nW, H, m_tok, d)
        v = v.view(BnW, m_tok, H, d).permute(0, 2, 1, 3).contiguous()   # (B*nW, H, m_tok, d)

        # attention logits
        attn = torch.matmul(q, k.transpose(-1, -2)) * self.scale  # (B*nW, H, n_tok, m_tok)

        # relative and CLS biases
        if self.attentionBias:
            rel = self.relative_position_bias_table[self.relative_position_index.view(-1)]
            rel = rel.view(N, M, H).permute(2, 0, 1).contiguous()  # (H, N, M)
            rel = rel.to(dtype=attn.dtype, device=attn.device)
            attn[:, :, 1:, 1:] = attn[:, :, 1:, 1:] + rel.unsqueeze(0)  # (1, H, N, M)
            attn[:, :, :1, :1] = attn[:, :, :1, :1] + self.cls_bias_self.to(attn.dtype)
            attn[:, :, :1, 1:] = attn[:, :, :1, 1:] + self.cls_bias_sequence_up.to(attn.dtype)
            attn[:, :, 1:, :1] = attn[:, :, 1:, :1] + self.cls_bias_sequence_down.to(attn.dtype)

        # masking first/last windows (broadcasted boolean masks, no .repeat)
        mask_left, mask_right = mask  # (maskCount, n_tok, m_tok), bool
        attn = attn.view(B, nW, H, n_tok, m_tok)
        maskCount = min(mask_left.shape[0], attn.shape[1])
        if maskCount > 0:
            neg_inf = max_neg_value(attn)
            attn[:, :maskCount].masked_fill_(mask_left[:maskCount].unsqueeze(0).unsqueeze(2), neg_inf)
            attn[:, -maskCount:].masked_fill_(mask_right[-maskCount:].unsqueeze(0).unsqueeze(2), neg_inf)
        attn = attn.view(BnW, H, n_tok, m_tok)

        # softmax, dropout
        attn = self.softmax(attn)
        attn = self.attnDrop(attn)

        # apply attention to values, merge heads, project
        out = torch.matmul(attn, v)                   # (B*nW, H, n_tok, d)
        out = out.permute(0, 2, 1, 3).contiguous()    # (B*nW, n_tok, H, d)
        out = out.view(BnW, n_tok, H * d)             # (B*nW, n_tok, H*d)
        out = self.proj(out)
        out = self.projDrop(out)
        return out


class FusedWindowTransformer(nn.Module):
    def __init__(
        self,
        dim,
        windowSize,
        shiftSize,
        receptiveSize,
        numHeads,
        headDim,
        mlpRatio,
        attentionBias,
        drop,
        attnDrop,
    ):
        super().__init__()
        self.attention = WindowAttention(
            dim=dim,
            windowSize=windowSize,
            receptiveSize=receptiveSize,
            numHeads=numHeads,
            headDim=headDim,
            attentionBias=attentionBias,
            attnDrop=attnDrop,
            projDrop=drop,
        )
        self.mlp = FeedForward(dim=dim, mult=mlpRatio, dropout=drop)
        self.attn_norm = nn.LayerNorm(dim)
        self.mlp_norm = nn.LayerNorm(dim)
        self.shiftSize = shiftSize

    def forward(self, x, cls, windowX, windowX_, mask, nW):
        """
        x : (B, T, C)
        cls : (B, nW, C)
        windowX  : (B*nW, 1+windowSize, C)
        windowX_ : (B*nW, 1+receptiveSize, C)
        mask : (mask_left, mask_right)
        nW : number of windows

        returns:
            xTrans  : (B, T, C)
            clsTrans: (B, nW, C)
        """
        # window attention over (B*nW, ·, ·)
        B, T, C = x.shape
        windowXTrans = self.attention(self.attn_norm(windowX), self.attn_norm(windowX_), mask, nW)
        clsTrans = windowXTrans[:, :1]   # (B*nW, 1, C)
        xTrans = windowXTrans[:, 1:]     # (B*nW, windowSize, C)

        clsTrans = clsTrans.view(B, nW, 1, C).squeeze(2)   # (B, nW, C)
        xTrans   = xTrans.view(B, nW, xTrans.size(1), C)   # (B, nW, windowSize, C)

        # overlap-add windows back into sequence
        xTrans = self.gatherWindows(xTrans, x.shape[1], self.shiftSize)

        # residual + MLP
        clsTrans = clsTrans + cls
        xTrans = xTrans + x
        xTrans = xTrans + self.mlp(self.mlp_norm(xTrans))
        clsTrans = clsTrans + self.mlp(self.mlp_norm(clsTrans))
        return xTrans, clsTrans

    def gatherWindows(self, windowedX, dynamicLength, shiftSize):
        """
        windowedX : (B, nW, windowLength, C)
        returns   : (B, dynamicLength, C)
        """
        B, nW, windowLength, C = windowedX.shape
        device = windowedX.device
        dtype = windowedX.dtype

        destination = torch.zeros((B, dynamicLength, C), device=device, dtype=dtype)
        scalerDestination = torch.zeros_like(destination)

        # indices of each window laid out on the time axis
        # shape: (nW, windowLength)
        idx = (
            torch.arange(windowLength, device=device).unsqueeze(0)
            + torch.arange(nW, device=device).unsqueeze(1) * shiftSize
        )
        idx = idx.view(1, nW, windowLength, 1).expand(B, nW, windowLength, C)  # (B, nW, windowLength, C)

        # Use reshape (safe for non-contiguous) instead of view
        src = windowedX.reshape(B, nW * windowLength, C)
        idx = idx.reshape(B, nW * windowLength, C)

        destination.scatter_add_(dim=1, index=idx, src=src)

        scalerSrc = torch.ones((B, nW * windowLength, C), device=device, dtype=dtype)
        scalerDestination.scatter_add_(dim=1, index=idx, src=scalerSrc)

        destination = destination / scalerDestination.clamp_min_(1)
        return destination


class BolTransformerBlock(nn.Module):
    def __init__(
        self,
        dim,
        numHeads,
        headDim,
        windowSize,
        receptiveSize,
        shiftSize,
        mlpRatio=1.0,
        drop=0.0,
        attnDrop=0.0,
        attentionBias=True,
    ):
        super().__init__()
        assert (receptiveSize - windowSize) % 2 == 0

        self.transformer = FusedWindowTransformer(
            dim=dim,
            windowSize=windowSize,
            shiftSize=shiftSize,
            receptiveSize=receptiveSize,
            numHeads=numHeads,
            headDim=headDim,
            mlpRatio=mlpRatio,
            attentionBias=attentionBias,
            drop=drop,
            attnDrop=attnDrop,
        )

        self.windowSize = windowSize
        self.receptiveSize = receptiveSize
        self.shiftSize = shiftSize
        self.remainder = (self.receptiveSize - self.windowSize) // 2

        # Precompute masks for non-matching query/key positions (boolean, broadcasted later)
        maskCount = self.remainder // shiftSize + 1
        mask_left = torch.zeros(maskCount, self.windowSize + 1, self.receptiveSize + 1, dtype=torch.bool)
        mask_right = torch.zeros_like(mask_left)
        for i in range(maskCount):
            if self.remainder > 0:
                mask_left[i, :, 1 : 1 + self.remainder - self.shiftSize * i] = True
                if (-self.remainder + self.shiftSize * i) > 0:
                    mask_right[maskCount - 1 - i, :, -self.remainder + self.shiftSize * i :] = True

        self.register_buffer("mask_left", mask_left)
        self.register_buffer("mask_right", mask_right)

    def forward(self, x, cls):
        """
        x   : (B, dynamicLength, C)
        cls : (B, nW, C)

        returns:
            fusedX_trans : (B, dynamicLength, C)
            cls_trans    : (B, nW, C)
        """
        B, _, C = x.shape
        device = x.device

        # Update Z to the exact covered length by windows
        Z = self.windowSize + self.shiftSize * (cls.shape[1] - 1)
        x = x[:, :Z]

        # Pad for receptive keys/values
        x_ = torch.cat(
            [torch.zeros((B, self.remainder, C), device=device), x, torch.zeros((B, self.remainder, C), device=device)],
            dim=1,
        )  # (B, remainder+Z+remainder, C)

        # Window the sequences (B, nW, C, window/receptive) -> transpose to (B, nW, window/receptive, C)
        windowedX, _ = windowBoldSignal(x.transpose(2, 1), self.windowSize, self.shiftSize)
        windowedX = windowedX.transpose(2, 3)

        windowedX_, _ = windowBoldSignal(x_.transpose(2, 1), self.receptiveSize, self.shiftSize)
        windowedX_ = windowedX_.transpose(2, 3)

        nW = windowedX.shape[1]

        # Insert CLS token as first position in each window; then flatten windows into batch
        xcls  = torch.cat([cls.unsqueeze(2),  windowedX],  dim=2)  # (B, nW, 1+W, C)
        xcls  = xcls.view(B * nW, xcls.size(2), C)                 # (B*nW, 1+W, C)
        xcls_ = torch.cat([cls.unsqueeze(2),  windowedX_], dim=2)  # (B, nW, 1+R, C)
        xcls_ = xcls_.view(B * nW, xcls_.size(2), C)               # (B*nW, 1+R, C)

        masks = (self.mask_left, self.mask_right)

        # Fused attention + MLP
        xTrans, clsTrans = self.transformer(x, cls, xcls, xcls_, masks, nW)
        return xTrans, clsTrans
